from __future__ import annotations

import itertools
from collections.abc import Callable
from dataclasses import dataclass
from datetime import date
from functools import reduce
from typing import Any, ClassVar

import geopandas as gpd
import pandas as pd
from dagster import AssetIn, AssetOut, MetadataValue, asset, multi_asset
from icecream import ic
from more_itertools import batched
from poppusher.assets.country import Country
from poppusher.cloud_outputs import (
    GeometryOutput,
    MetricsOutput,
    send_to_metrics_metadata_sensor,
)
from poppusher.metadata import (
    COL,
    CountryMetadata,
    DataPublisher,
    GeometryMetadata,
    MetricMetadata,
    SourceDataRelease,
    metadata_to_dataframe,
)
from poppusher.utils import add_metadata

from .census_tasks import (
    ACS_METADATA,
    SUMMARY_LEVELS,
    extract_values_at_specified_levels,
    generate_variable_dictionary,
    get_geom_ids_table_for_summary,
    get_summary_table,
    get_summary_table_file_names,
    select_estimates,
)

SUMMARY_LEVEL_STRINGS = ["oneYear", "fiveYear"]
GEOMETRY_COL = "AFFGEOID"
METRICS_COL = "GEO_ID"

# For testing
# REQUIRED_TABLES = [
#     "acsdt1y2019-b01001.dat",
#     "acsdt1y2019-b01001.dat",
#     "acsdt1y2019-b01001a.dat",
#     "acsdt1y2019-b01001b.dat",
#     "acsdt1y2019-b01001d.dat",
# ]
# BATCH_SIZE = 2

# For prod
REQUIRED_TABLES = None
BATCH_SIZE = 10


@dataclass
class DerivedColumn:
    hxltag: str
    extended_variable_names: list[str]
    filter_func: Callable[[pd.DataFrame, list[str]], pd.DataFrame]
    output_column_name: str
    human_readable_name: str
    table_id: str


INFANTS = ["Total|Male|Under 5 years", "Total|Female|Under 5 years"]
CHILDREN = [
    "Total|Male|Under 5 years",
    "Total|Male|5 to 9 years",
    "Total|Male|10 to 14 years",
    "Total|Male|15 to 17 years",
    "Total|Female|Under 5 years",
    "Total|Female|5 to 9 years",
    "Total|Female|10 to 14 years",
    "Total|Female|15 to 17 years",
]
CHILDREN_5_TO_17 = [
    "Total|Male|5 to 9 years",
    "Total|Male|10 to 14 years",
    "Total|Male|15 to 17 years",
    "Total|Female|5 to 9 years",
    "Total|Female|10 to 14 years",
    "Total|Female|15 to 17 years",
]
ADULTS_MALE = [
    "Total|Male|18 and 19 years",
    "Total|Male|20 years",
    "Total|Male|21 years",
    "Total|Male|22 to 24 years",
    "Total|Male|25 to 29 years",
    "Total|Male|30 to 34 years",
    "Total|Male|35 to 39 years",
    "Total|Male|40 to 44 years",
    "Total|Male|45 to 49 years",
    "Total|Male|50 to 54 years",
    "Total|Male|55 to 59 years",
    "Total|Male|60 and 61 years",
    "Total|Male|62 to 64 years",
    "Total|Male|65 and 66 years",
    "Total|Male|67 to 69 years",
    "Total|Male|70 to 74 years",
    "Total|Male|75 to 79 years",
    "Total|Male|80 to 84 years",
    "Total|Male|85 years and over",
]
ADULTS_FEMALE = [
    "Total|Female|18 and 19 years",
    "Total|Female|20 years",
    "Total|Female|21 years",
    "Total|Female|22 to 24 years",
    "Total|Female|25 to 29 years",
    "Total|Female|30 to 34 years",
    "Total|Female|35 to 39 years",
    "Total|Female|40 to 44 years",
    "Total|Female|45 to 49 years",
    "Total|Female|50 to 54 years",
    "Total|Female|55 to 59 years",
    "Total|Female|60 and 61 years",
    "Total|Female|62 to 64 years",
    "Total|Female|65 and 66 years",
    "Total|Female|67 to 69 years",
    "Total|Female|70 to 74 years",
    "Total|Female|75 to 79 years",
    "Total|Female|80 to 84 years",
    "Total|Female|85 years and over",
]
ADULTS = ADULTS_MALE + ADULTS_FEMALE
INDIVIDUALS = ["Total"]


# Config for each partition to be derived
age_code = "`Age Code`"
sex_label = "`Sex Label`"
DERIVED_COLUMNS = [
    DerivedColumn(
        hxltag="#population+children+age5_17",
        filter_func=lambda df, cols: df[cols].sum(axis=1).to_frame(),
        extended_variable_names=CHILDREN_5_TO_17,
        output_column_name="children_5_17",
        human_readable_name="Children aged 5 to 17",
        table_id="B01001",
    ),
    DerivedColumn(
        hxltag="#population+infants+age0_4",
        extended_variable_names=INFANTS,
        filter_func=lambda df, cols: df[cols].sum(axis=1).to_frame(),
        output_column_name="infants_0_4",
        human_readable_name="Infants aged 0 to 4",
        table_id="B01001",
    ),
    DerivedColumn(
        hxltag="#population+children+age0_17",
        extended_variable_names=CHILDREN,
        filter_func=lambda df, cols: df[cols].sum(axis=1).to_frame(),
        output_column_name="children_0_17",
        human_readable_name="Children aged 0 to 17",
        table_id="B01001",
    ),
    DerivedColumn(
        hxltag="#population+adults+f",
        extended_variable_names=ADULTS_FEMALE,
        filter_func=lambda df, cols: df[cols].sum(axis=1).to_frame(),
        output_column_name="adults_f",
        human_readable_name="Female adults",
        table_id="B01001",
    ),
    DerivedColumn(
        hxltag="#population+adults+m",
        extended_variable_names=ADULTS_MALE,
        filter_func=lambda df, cols: df[cols].sum(axis=1).to_frame(),
        output_column_name="adults_m",
        human_readable_name="Male adults",
        table_id="B01001",
    ),
    DerivedColumn(
        hxltag="#population+adults",
        extended_variable_names=ADULTS,
        filter_func=lambda df, cols: df[cols].sum(axis=1).to_frame(),
        output_column_name="adults",
        human_readable_name="Adults",
        table_id="B01001",
    ),
    DerivedColumn(
        hxltag="#population+ind",
        extended_variable_names=INDIVIDUALS,
        filter_func=lambda df, cols: df[cols].sum(axis=1).to_frame(),
        output_column_name="individuals",
        human_readable_name="Total individuals",
        table_id="B01001",
    ),
]

DERIVED_COLUMN_SPECIFICATIONS: dict[str, list[DerivedColumn]] = {
    f"{year}/{summary_level}/{geo_level}/0": DERIVED_COLUMNS
    for year, summary_level, geo_level in itertools.product(
        ["2019", "2020", "2021"],
        ["oneYear", "fiveYear"],
        ["tract", "block_group", "county"],
    )
}


class USA(Country):
    country_metadata: ClassVar[CountryMetadata] = CountryMetadata(
        name_short_en="United States",
        name_official="United States of America",
        iso2="US",
        iso3="USA",
        iso3166_2=None,
    )
    required_tables: list[str] | None = None

    def _country_metadata(self, _context) -> CountryMetadata:
        return self.country_metadata

    def _data_publisher(
        self, _context, country_metadata: CountryMetadata
    ) -> DataPublisher:
        return DataPublisher(
            name="United States Census Bureau",
            url="https://www.census.gov/programs-surveys/acs",
            description=(
                """
                The United States Census Bureau, officially the Bureau of the Census,
                is a principal agency of the U.S. Federal Statistical System, responsible
                for producing data about the American people and economy.
                """
            ),
            countries_of_interest=[country_metadata.id],
        )

    def _catalog(self, context) -> pd.DataFrame:
        self.remove_all_partition_keys(context)
        catalog_list = []
        for year, _ in ACS_METADATA.items():
            for summary_level in SUMMARY_LEVEL_STRINGS:
                for geo_level in SUMMARY_LEVELS:
                    # If year and summary level has no data, skip it.
                    if (ACS_METADATA[year][summary_level] is None) or (
                        summary_level == "oneYear"
                        and (geo_level == "tract" or geo_level == "block_group")
                    ):
                        continue

                    table_names_list = [
                        table_name
                        for table_name in get_summary_table_file_names(
                            year, summary_level
                        )
                        if (
                            REQUIRED_TABLES is not None
                            and table_name in REQUIRED_TABLES
                        )
                        or REQUIRED_TABLES is None
                    ]

                    table_names_list = list(batched(table_names_list, BATCH_SIZE))

                    # Catalog
                    table_names = pd.DataFrame({"table_names_batch": table_names_list})
                    table_names["year"] = year
                    table_names["summary_level"] = summary_level
                    table_names["geo_level"] = geo_level
                    table_names["batch"] = range(len(table_names_list))
                    table_names["partition_key"] = (
                        table_names["year"].astype(str)
                        + "/"
                        + table_names["summary_level"].astype(str)
                        + "/"
                        + table_names["geo_level"].astype(str)
                        + "/"
                        + table_names["batch"].astype(str)
                        # .apply(lambda x: x.split(".")[0])
                    )
                    catalog_list.append(table_names)

        catalog = pd.concat(catalog_list, axis=0).reset_index(drop=True)
        self.add_partition_keys(context, catalog["partition_key"].to_list())
        add_metadata(context, catalog, "Catalog")
        return catalog

    def _geometry(
        self, context
    ) -> list[tuple[GeometryMetadata, gpd.GeoDataFrame, pd.DataFrame]]:
        geometries_to_return = []
        for year, metadata in ACS_METADATA.items():
            context.log.debug(ic(year))
            context.log.debug(ic(metadata))
            names_col = ACS_METADATA[year]["geoIdCol"]
            # Combine fiveYear and oneYear geoIDs
            if year != 2020:
                geo_ids5 = get_geom_ids_table_for_summary(year, "fiveYear")
                geo_ids1 = get_geom_ids_table_for_summary(year, "oneYear")
                geo_ids = (
                    pd.concat(
                        [
                            geo_ids5,
                            geo_ids1[~geo_ids1[names_col].isin(geo_ids5[names_col])],
                        ],
                        axis=0,
                    )
                    .reset_index(drop=True)
                    .rename(columns={names_col: "GEO_ID", "NAME": "eng"})
                )
            else:
                geo_ids = (
                    get_geom_ids_table_for_summary(year, "fiveYear")
                    .rename(columns={names_col: "GEO_ID", "NAME": "eng"})
                    .loc[:, ["GEO_ID", "eng"]]
                )

            for geo_level, url in metadata["geoms"].items():
                geometry_metadata = GeometryMetadata(
                    country_metadata=self.country_metadata,
                    validity_period_start=date(year, 1, 1),
                    validity_period_end=date(year, 1, 1),
                    level=geo_level,
                    # TODO: what should hxl_tag be?
                    hxl_tag=geo_level,
                )

                region_geometries_raw: gpd.GeoDataFrame = gpd.read_file(url)

                # Copy names
                region_names = geo_ids.copy()
                region_geometries_raw = region_geometries_raw.dissolve(
                    by=GEOMETRY_COL
                ).reset_index()

                context.log.debug(ic(region_geometries_raw.head()))
                context.log.debug(ic(region_geometries_raw.columns))
                region_geometries = region_geometries_raw.rename(
                    columns={GEOMETRY_COL: "GEO_ID"}
                ).loc[:, ["geometry", "GEO_ID"]]
                context.log.debug(ic(region_geometries.head()))
                context.log.debug(ic(region_geometries.columns))

                # TODO: Merge names.
                # TODO: Check this step. Is this subsetting giving the correct GEO_IDs?
                region_geometries = region_geometries.loc[
                    region_geometries["GEO_ID"].isin(region_names["GEO_ID"])
                ]

                # Rename cols for names
                region_names = region_names.loc[
                    # Subset to row in geoms file
                    region_names["GEO_ID"].isin(region_geometries["GEO_ID"]),
                    ["GEO_ID", "eng"],
                ].drop_duplicates()
                context.log.debug(ic(region_names.head()))
                context.log.debug(ic(region_names.columns))

                geometries_to_return.append(
                    GeometryOutput(
                        metadata=geometry_metadata,
                        gdf=region_geometries,
                        names_df=region_names,
                    )
                )

        return geometries_to_return

    def _source_data_releases(
        self, _context, geometry, data_publisher
    ) -> dict[str, SourceDataRelease]:
        source_data_releases = {}

        idx = 0
        for year, metadata in ACS_METADATA.items():
            for _, url in metadata["geoms"].items():
                for summary_level in ["oneYear", "fiveYear"]:
                    geo = geometry[idx]
                    source_data_release: SourceDataRelease = SourceDataRelease(
                        name=(
                            f"ACS {year} 1 year"
                            if summary_level == "oneYear"
                            else f"ACS {year} 5 year"
                        ),
                        date_published=date(year, 1, 1),
                        reference_period_start=date(year, 1, 1),
                        reference_period_end=date(year, 1, 1),
                        collection_period_start=date(year, 1, 1),
                        collection_period_end=date(year, 1, 1),
                        expect_next_update=date(year, 1, 1),
                        url=url,
                        data_publisher_id=data_publisher.id,
                        description="""
                            The American Community Survey (ACS) helps local officials,
                            community leaders, and businesses understand the changes
                            taking place in their communities. It is the premier source
                            for detailed population and housing information about our nation.
                        """,
                        geometry_metadata_id=geo.metadata.id,
                    )
                    source_data_releases[
                        f"{year}_{summary_level}_{geo.metadata.level}"
                    ] = source_data_release
                idx += 1

        return source_data_releases

    def _census_tables(self, context, catalog) -> pd.DataFrame:
        partition = context.asset_partition_key_for_output()
        ic(partition)
        ic(catalog.loc[catalog["partition_key"].eq(partition), "table_names_batch"])
        row = catalog.loc[catalog["partition_key"].eq(partition), :]
        table_names_batch = row.iloc[0]["table_names_batch"]
        year = row.iloc[0].loc["year"]
        summary_level = row.iloc[0].loc["summary_level"]
        geo_level = row.iloc[0].loc["geo_level"]

        # TODO: generate as an asset to cache result per year and summary_level
        geoids = get_geom_ids_table_for_summary(year, summary_level)
        census_tables = []
        for table_name in table_names_batch:
            df = get_summary_table(table_name, year, summary_level)
            values = extract_values_at_specified_levels(
                df, geoids, ACS_METADATA[int(year)]["geoIdCol"]
            )
            try:
                table = values[geo_level]
                context.log.info(ic(table))
                context.log.info(ic(table.columns))
                census_tables.append(table)
            except Exception as err:
                msg = (
                    f"Could not get table ({table_name}) at geo level ({geo_level}) "
                    f"for summary level ({summary_level}) in year ({year}) with "
                    f"error: {err}"
                )
                context.log.warning(msg)

        if len(census_tables) > 0:
            census_tables = reduce(
                lambda left, right: left.merge(
                    right, on=METRICS_COL, how="outer", validate="one_to_one"
                ),
                census_tables,
            )
        else:
            census_tables = pd.DataFrame()
            msg = (
                f"No tables at geo level ({geo_level}) "
                f"for summary level ({summary_level}) in year ({year})."
            )
            context.log.warning(msg)

        add_metadata(
            context, census_tables, title=context.asset_partition_key_for_output()
        )
        return census_tables

    def _source_metric_metadata(
        self,
        context,
        catalog: pd.DataFrame,
        source_data_releases: dict[str, SourceDataRelease],
    ) -> MetricMetadata:
        ...

    def _gen_url(
        self, col: str, table_names: list[str], year: int, summary_level: str
    ) -> str:
        base = ACS_METADATA[year]["base"]
        summary_file_dir = base + ACS_METADATA[year][summary_level]["tables"]
        for table_name in table_names:
            table_id = table_name.split("-")[1].split(".")[0].upper()
            col_start = col.split("_")[0]
            if col_start == table_id:
                return f"{summary_file_dir}/{table_name}"
        # If no URL can be generated, return "TBD"
        return "TBD"

    def _gen_parquet_path(self, partition_key: str) -> str:
        return "/".join(
            [
                self.key_prefix,
                "metrics",
                f"{''.join(c for c in partition_key if c.isalnum()) + '.parquet'}",
            ]
        )

    @staticmethod
    def _column_to_variable(name: str) -> str:
        split = name.split("_")
        return split[0] + "_" + split[1][1:]

    @staticmethod
    def _variable_to_column(variable: str, type: str = "M") -> str:
        split = variable.split("_")
        return split[0] + f"_{type}" + split[1]

    def make_partial_metric_metadata(
        self,
        column: str,
        variable_dictionary: pd.DataFrame,
        source_data_release: SourceDataRelease,
        partition_key: str,
        table_names: Any,
        year: str,
        summary_level: str,
    ) -> MetricMetadata:
        variable = self._column_to_variable(column)
        info = (
            variable_dictionary.loc[variable_dictionary["uniqueID"].eq(variable)]
            .iloc[0]
            .to_dict()
        )

        def gen_description(info: dict[str, str]) -> str:
            return "; ".join(
                [f"Key: {key}, Value: {value}" for key, value in info.items()]
            )

        def gen_hxl_tag(info: dict[str, str]) -> str:
            return (
                "#"
                + "".join(
                    [c for c in info["universe"].title() if c != " " and c.isalnum()]
                )
                + "+"
                + "+".join(
                    "".join(c for c in split.title() if c.isalnum() and c != " ")
                    for split in info["variableExtendedName"].split("|")
                )
            )

        def gen_human_readable_name() -> str:
            return (
                f"{info['universe']}, {info['tableName']}, {year}, {summary_level}, "
                f"{info['variableName']}"
            )

        return MetricMetadata(
            human_readable_name=gen_human_readable_name(),
            description=gen_description(info),
            hxl_tag=gen_hxl_tag(info),
            metric_parquet_path=self._gen_parquet_path(partition_key),
            parquet_column_name=column,
            parquet_margin_of_error_column=self._variable_to_column(variable, "E"),
            parquet_margin_of_error_file=self._variable_to_column(variable, "M"),
            potential_denominator_ids=None,
            # TODO: get value
            source_metric_id="TBD",
            parent_metric_id=None,
            source_data_release_id=source_data_release.id,
            # TODO: check this works
            source_download_url=self._gen_url(
                column, table_names, int(year), summary_level
            ),
            source_archive_file_path=None,
            # TODO: get value
            source_documentation_url="TBD",
        )

    def create_derived_metrics(self):
        """
        Creates an asset providing the metrics derived from the census tables and the
        corresponding source metric metadata.
        """

        @multi_asset(
            partitions_def=self.dataset_node_partition,
            ins={
                "catalog": AssetIn(key_prefix=self.key_prefix),
                "census_tables": AssetIn(key_prefix=self.key_prefix),
                "source_data_releases": AssetIn(key_prefix=self.key_prefix),
            },
            outs={
                "metrics_metadata_partitioned": AssetOut(key_prefix=self.key_prefix),
                "metrics_partitioned": AssetOut(key_prefix=self.key_prefix),
            },
        )
        def derived_metrics(
            context,
            catalog: pd.DataFrame,
            census_tables: pd.DataFrame,
            source_data_releases: dict[str, SourceDataRelease],
        ) -> tuple[list[MetricMetadata], MetricsOutput]:
            partition_key = context.partition_key

            row = catalog[catalog["partition_key"] == partition_key].iloc[0].to_dict()
            year = row["year"]
            summary_level = row["summary_level"]
            geo_level = row["geo_level"]
            table_names = row["table_names_batch"]

            # TODO: consider refactoring as asset
            variable_dictionary = generate_variable_dictionary(year, summary_level)
            if census_tables.shape[0] == 0 or census_tables.shape[1] == 0:
                context.log.warning(f"No data found in parition: {partition_key}")
                return [], MetricsOutput(metadata=[], metrics=pd.DataFrame())

            metrics = census_tables.copy().set_index(METRICS_COL)

            if metrics.shape[1] == 0:
                context.log.warning(f"No metrics found in parition: {partition_key}")
                return [], MetricsOutput(metadata=[], metrics=pd.DataFrame())

            estimates = select_estimates(metrics)
            # TODO: No need to select errors, unless to check there is an error column
            # errors = select_errors(metrics)

            if estimates.shape[1] == 0:
                context.log.warning(f"No estimates found in parition: {partition_key}")
                return [], MetricsOutput(metadata=[], metrics=pd.DataFrame())

            derived_mmd = []
            metrics_out = []
            # Construct derived metrics if any
            try:
                # Make copy for derived estimates
                for derived_column in DERIVED_COLUMN_SPECIFICATIONS[partition_key]:
                    # Log derived column
                    ic(context.log.debug(derived_column))
                    derived = estimates.copy()
                    # Get uniqueID for metrics of the extended variable name
                    # from dictionary
                    cols = variable_dictionary[
                        variable_dictionary["variableExtendedName"].isin(
                            derived_column.extended_variable_names
                        )
                        & variable_dictionary["universe"]
                        .str.lower()
                        .eq("total population")
                        & variable_dictionary["tableID"].str.startswith(
                            derived_column.table_id
                        )
                    ]["uniqueID"].to_list()
                    assert len(set(cols)) == len(
                        set(derived_column.extended_variable_names)
                    )

                    # Convert metric table column names to variable names
                    derived = derived.rename(
                        columns={
                            col: self._column_to_variable(col)
                            for col in derived.columns
                        }
                    )
                    derived_metric_out = (
                        derived[cols]
                        .sum(axis=1)
                        .to_frame()
                        .rename(columns={0: derived_column.output_column_name})
                        .reset_index()
                    )
                    source_data_release = source_data_releases[
                        f"{year}_{summary_level}_{geo_level}"
                    ]
                    derived_metric_metadata_out = MetricMetadata(
                        human_readable_name=derived_column.human_readable_name,
                        description=derived_column.human_readable_name,
                        hxl_tag=derived_column.hxltag,
                        metric_parquet_path=self._gen_parquet_path(partition_key),
                        parquet_column_name=derived_column.output_column_name,
                        # TODO: add error from combination of columns
                        parquet_margin_of_error_column=None,
                        parquet_margin_of_error_file=None,
                        potential_denominator_ids=None,
                        # TODO: get value
                        source_metric_id="TBD",
                        parent_metric_id=None,
                        source_data_release_id=source_data_release.id,
                        # TODO: check this works
                        source_download_url=self._gen_url(
                            cols[0], table_names, int(year), summary_level
                        ),
                        source_archive_file_path=None,
                        # TODO: get value
                        source_documentation_url="TBD",
                    )
                    metrics_out.append(derived_metric_out)
                    derived_mmd.append(derived_metric_metadata_out)
                    # Log results
                    ic(context.log.debug(derived_metric_out))
                    ic(context.log.debug(derived_metric_metadata_out))

            except:
                context.log.debug(
                    f"Partition key ({partition_key}) has no additional derived metric specifications."
                )

            # Add remaining metrics
            for col in estimates.columns:
                metric_metadata = self.make_partial_metric_metadata(
                    col,
                    variable_dictionary,
                    source_data_releases[f"{year}_{summary_level}_{geo_level}"],
                    partition_key,
                    table_names,
                    year,
                    summary_level,
                )
                derived_mmd.append(metric_metadata)

            metrics_out.append(metrics)
            metrics = reduce(
                lambda left, right: left.merge(
                    right, on=COL.GEO_ID.value, how="inner", validate="one_to_one"
                ),
                metrics_out,
            )

            context.add_output_metadata(
                output_name="metrics_metadata_partitioned",
                metadata={
                    "metadata_preview": MetadataValue.md(
                        metadata_to_dataframe(derived_mmd).head().to_markdown()
                    ),
                },
            )
            context.add_output_metadata(
                output_name="metrics_partitioned",
                metadata={
                    "metadata_preview": MetadataValue.md(
                        metadata_to_dataframe(derived_mmd).head().to_markdown()
                    ),
                    "metrics_shape": f"{metrics.shape[0]} rows x {metrics.shape[1]} columns",
                    "metrics_preview": MetadataValue.md(metrics.head().to_markdown()),
                },
            )
            return derived_mmd, MetricsOutput(metadata=derived_mmd, metrics=metrics)

        return derived_metrics

    # Implementation not required since overridden
    def _derived_metrics(self, _census_tables):
        ...

    def create_metrics_metadata_output(self):
        @send_to_metrics_metadata_sensor
        @asset(key_prefix=self.key_prefix)
        def metrics_metadata_output(
            context,
            metrics_metadata_partitioned,
        ) -> list[MetricMetadata]:
            partition_names = context.instance.get_dynamic_partitions(
                self.partition_name
            )
            if len(partition_names) == 1:
                metrics_metadata_partitioned = {
                    partition_names[0]: metrics_metadata_partitioned
                }

            outputs = [
                mmd
                for output in metrics_metadata_partitioned.values()
                if len(metrics_metadata_partitioned) > 0
                for mmd in output
            ]
            context.add_output_metadata(
                metadata={
                    # TODO: check values are correct
                    "num_metrics": len(outputs),
                    "num_parquets": len(partition_names),
                },
            )
            return outputs

        return metrics_metadata_output

    def create_metrics_partitioned_output(self):
        # TODO: need to implement partitioned assets to be found by sensor
        # For now use IO manager directly
        # @send_to_metrics_partitioned_sensor
        @asset(
            key_prefix=self.key_prefix,
            partitions_def=self.dataset_node_partition,
            io_manager_key="metrics_partitioned_io_manager",
        )
        def metrics_partitioned_output(
            context,
            metrics_partitioned: MetricsOutput,
        ) -> MetricsOutput:
            metadata, metrics = (
                metrics_partitioned.metadata,
                metrics_partitioned.metrics,
            )
            if len(metadata) == 0:
                err_msg = (
                    f"No metrics output for partition key: {context.partition_key}"
                )
                context.log.warning(err_msg)
                return metrics_partitioned

            context.add_output_metadata(
                metadata={
                    "metadata_preview": MetadataValue.md(
                        metadata_to_dataframe(metadata).head().to_markdown()
                    ),
                    "metrics_shape": f"{metrics.shape[0]} rows x {metrics.shape[1]} columns",
                    "metrics_preview": MetadataValue.md(metrics.head().to_markdown()),
                },
            )
            return metrics_partitioned

        return metrics_partitioned_output


# Assets
usa = USA()
country_metadata = usa.create_country_metadata()
data_publisher = usa.create_data_publisher()
geometry = usa.create_geometry()
source_data_releases = usa.create_source_data_releases()
catalog = usa.create_catalog()
census_tables = usa.create_census_tables()
derived_metrics = usa.create_derived_metrics()
metrics_metadata_output = usa.create_metrics_metadata_output()
metrics_partitioned_output = usa.create_metrics_partitioned_output()



from __future__ import annotations

import os
import re
import zipfile
from collections.abc import Callable, Iterable
from dataclasses import dataclass
from datetime import date
from functools import partial
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, ClassVar, Literal
from urllib.parse import urljoin

import dagster
import geopandas as gpd
import pandas as pd
import requests
from bs4 import BeautifulSoup
from dagster import MetadataValue
from icecream import ic
from pandas import DataFrame

from poppusher.assets.country import Country
from poppusher.cloud_outputs import GeometryOutput, MetricsOutput
from poppusher.metadata import (
    COL,
    CountryMetadata,
    DataPublisher,
    GeometryMetadata,
    MetricMetadata,
    SourceDataRelease,
    metadata_to_dataframe,
)
from poppusher.utils import (
    SourceDataAssumptionsOutdated,
    add_metadata,
    extract_main_file_from_zip,
    markdown_from_plot,
)

# Overview
# - Create a asset which is a catalog of the available data / tables / metrics
# - The available geometry levels are only discoverable after downloading the zip file
# - The zip files can contain multiple CSV files, one for each geometry level
# - Some of the downloaded files mistakenly have two consecutive `.` in the filename, e.g. `census2021-ts002-lsoa..csv`. We need to be able to gracefully handle this
# - The catalog must to parsed into an Dagster Partition, so that
#    - individual tables can be uploaded to the cloud table sensor
#    - the metadata object can be created for each table/metric


@dataclass
class EWCensusGeometryLevel:
    level: str
    geo_id_column: str
    name_columns: dict[str, str]  # keys = language codes, values = column names
    data_download_url: str
    documentation_url: str
    hxl_tag: str = ""

    def __post_init__(self):
        if self.hxl_tag == "":
            self.hxl_tag = f"#geo+bounds+code+{self.level}"


@dataclass
class SourceTable:
    hxltag: str
    geo_level: str
    geo_column: str
    source_column: str


@dataclass
class DerivedColumn:
    hxltag: str
    # If `None`, then just the named `source_column` will be used
    column_select: Callable[[pd.DataFrame], list[str]]
    output_column_name: str
    human_readable_name: str


def census_table_metadata(
    catalog_row: dict[str, str],
    source_table: SourceTable,
    source_data_releases: dict[str, SourceDataRelease],
) -> MetricMetadata:
    return MetricMetadata(
        human_readable_name=catalog_row["human_readable_name"],
        source_download_url=catalog_row["source_download_url"],
        source_archive_file_path=catalog_row["source_archive_file_path"],
        source_documentation_url=catalog_row["source_documentation_url"],
        source_data_release_id=source_data_releases[source_table.geo_level].id,
        # TODO - this is a placeholder
        parent_metric_id="unknown_at_this_stage",
        potential_denominator_ids=None,
        parquet_margin_of_error_file=None,
        parquet_margin_of_error_column=None,
        parquet_column_name=source_table.source_column,
        # TODO - this is a placeholder
        metric_parquet_path="unknown_at_this_stage",
        hxl_tag=source_table.hxltag,
        description=catalog_row["description"],
        source_metric_id=source_table.hxltag,
    )


CENSUS_COLLECTION_DATE = date(2021, 3, 21)

EW_CENSUS_GEO_LEVELS: dict[str, EWCensusGeometryLevel] = {
    "oa": EWCensusGeometryLevel(
        level="oa",
        geo_id_column="oa21cd",
        name_columns={"eng": "name"},
        data_download_url="https://borders.ukdataservice.ac.uk/ukborders/easy_download/prebuilt/shape/Ew_oa_2021.zip",
        documentation_url="https://borders.ukdataservice.ac.uk/easy_download_data.html?data=Ew_oa_2021",
    ),
    "lsoa": EWCensusGeometryLevel(
        level="lsoa",
        geo_id_column="lsoa21cd",
        name_columns={"eng": "name"},
        data_download_url="https://borders.ukdataservice.ac.uk/ukborders/easy_download/prebuilt/shape/Ew_lsoa_2021.zip",
        documentation_url="https://borders.ukdataservice.ac.uk/easy_download_data.html?data=Ew_lsoa_2021",
    ),
    "msoa": EWCensusGeometryLevel(
        level="msoa",
        geo_id_column="msoa21cd",
        name_columns={"eng": "name"},
        data_download_url="https://borders.ukdataservice.ac.uk/ukborders/easy_download/prebuilt/shape/Ew_msoa_2021.zip",
        documentation_url="https://borders.ukdataservice.ac.uk/easy_download_data.html?data=Ew_msoa_2021",
    ),
    "ltla": EWCensusGeometryLevel(
        level="ltla",
        geo_id_column="ltla22cd",
        name_columns={"eng": "ltla22nm", "cym": "ltla22nmw"},
        data_download_url="https://borders.ukdataservice.ac.uk/ukborders/easy_download/prebuilt/shape/Ew_ltla_2022.zip",
        documentation_url="https://borders.ukdataservice.ac.uk/easy_download_data.html?data=Ew_ltla_2022",
    ),
    "rgn": EWCensusGeometryLevel(
        level="rgn",
        geo_id_column="rgn22cd",
        name_columns={"eng": "rgn22nm", "cym": "rgn22nmw"},
        data_download_url="https://borders.ukdataservice.ac.uk/ukborders/easy_download/prebuilt/shape/Ew_rgn_2022.zip",
        documentation_url="https://borders.ukdataservice.ac.uk/easy_download_data.html?data=Ew_rgn_2022",
    ),
    "ctry": EWCensusGeometryLevel(
        level="ctry",
        geo_id_column="ctry22cd",
        name_columns={"eng": "ctry22nm", "cym": "ctry22nmw"},
        data_download_url="https://borders.ukdataservice.ac.uk/ukborders/easy_download/prebuilt/shape/Ew_ctry_2022.zip",
        documentation_url="https://borders.ukdataservice.ac.uk/easy_download_data.html?data=Ew_ctry_2022",
    ),
}


# TODO - this is probably only required for tests,
# hence would be best move to a test fixture
REQUIRED_TABLES = ["TS009"] if os.getenv("ENV") == "dev" else None


# TODO - these regexes are probably only useful for table TS009.
# At present that is the only table we use using for any of the derived metrics
# In future we might need to structure these in a more scalable manner
SexCategory = Literal["female", "male", "total"]

sex_regexes: dict[SexCategory, re.Pattern[str]] = {
    "total": re.compile(r"Sex: All persons;"),
    "female": re.compile(r"Sex: Female;"),
    "male": re.compile(r"Sex: Male;"),
}

age_regex = re.compile(
    r"Age: (Aged (?P<age>\d\d?) years?;|(?P<baby>Aged under 1 year);|(?P<ninetyplus>Aged 90 years and over);) measures: Value"
)
all_ages_regex = re.compile(r"Age: Total; measures: Value")


def columns_selector(
    columns_list: Iterable[Any], age_range: list[int] | None, sex: SexCategory
):
    sex_regex = sex_regexes[sex]

    columns_to_sum = []
    for col in columns_list:
        if not sex_regex.search(col):
            continue

        # No age range specified, so just look for the total column
        if age_range is None:
            match = all_ages_regex.search(col)
            if match:
                columns_to_sum.append(col)
            continue

        # Else look for the specific age range
        match = age_regex.search(col)
        if match and (
            (match.group("baby") and 0 in age_range)
            or (match.group("ninetyplus") and 90 in age_range)
            or (match.group("age") and int(match.group("age")) in age_range)
        ):
            columns_to_sum.append(col)

    return columns_to_sum


DERIVED_COLUMNS = [
    DerivedColumn(
        hxltag="#population+children+age5_17",
        column_select=lambda df: columns_selector(
            df.columns, list(range(5, 18)), "total"
        ),
        output_column_name="children_5_17",
        human_readable_name="Children aged 5 to 17",
    ),
    DerivedColumn(
        hxltag="#population+infants+age0_4",
        column_select=lambda df: columns_selector(
            df.columns,
            list(range(5)),
            "total",
        ),
        output_column_name="infants_0_4",
        human_readable_name="Infants aged 0 to 4",
    ),
    DerivedColumn(
        hxltag="#population+children+age0_17",
        column_select=lambda df: columns_selector(df.columns, list(range(18)), "total"),
        output_column_name="children_0_17",
        human_readable_name="Children aged 0 to 17",
    ),
    DerivedColumn(
        hxltag="#population+adults+f",
        output_column_name="adults_f",
        column_select=lambda df: columns_selector(
            df.columns, list(range(18, 91)), "female"
        ),
        human_readable_name="Female adults",
    ),
    DerivedColumn(
        hxltag="#population+adults+m",
        column_select=lambda df: columns_selector(
            df.columns, list(range(18, 91)), "male"
        ),
        output_column_name="adults_m",
        human_readable_name="Male adults",
    ),
    DerivedColumn(
        hxltag="#population+adults",
        column_select=lambda df: columns_selector(
            df.columns, list(range(18, 91)), "total"
        ),
        output_column_name="adults",
        human_readable_name="Adults",
    ),
    DerivedColumn(
        hxltag="#population+ind",
        column_select=lambda df: columns_selector(df.columns, None, "total"),
        output_column_name="individuals",
        human_readable_name="Total individuals",
    ),
]

# Lookup of `partition_key` (eg geom + source table id) to `DerivedColumn` (columns that can be derived from the source table)
DERIVED_COLUMN_SPECIFICATIONS: dict[str, list[DerivedColumn]] = {
    "ltla/TS009": DERIVED_COLUMNS,
}


class EnglandAndWales(Country):
    geo_levels: ClassVar[list[str]] = list(EW_CENSUS_GEO_LEVELS.keys())
    required_tables: list[str] | None = REQUIRED_TABLES
    country_metadata: ClassVar[CountryMetadata] = CountryMetadata(
        name_short_en="England and Wales",
        name_official="England and Wales",
        iso3="GBR",
        iso2="GB",
        iso3166_2="GB-EAW",
    )

    def _data_publisher(
        self, _context, _country_metdata: CountryMetadata
    ) -> DataPublisher:
        return DataPublisher(
            name="Office for National Statistics",
            url="https://www.nomisweb.co.uk/sources/census_2021_bulk",
            description="We are the UK's largest independent producer of official statistics and its recognised national statistical institute. We are responsible for collecting and publishing statistics related to the economy, population and society at national, regional and local levels. We also conduct the census in England and Wales every 10 years.",
            countries_of_interest=[self.country_metadata.id],
        )

    def _catalog(self, context) -> pd.DataFrame:
        self.remove_all_partition_keys(context)

        catalog_summary = {
            "node": [],
            "partition_key": [],
            "table_id": [],
            "geo_level": [],
            "human_readable_name": [],
            "description": [],
            "metric_parquet_file_url": [],
            "parquet_column_name": [],
            "parquet_margin_of_error_column": [],
            "parquet_margin_of_error_file": [],
            "potential_denominator_ids": [],
            "parent_metric_id": [],
            "source_data_release_id": [],
            "source_download_url": [],
            "source_format": [],
            "source_archive_file_path": [],
            "source_documentation_url": [],
        }

        bulk_downloads_df = bulk_downloads_webpage()

        for bulk_downloads_index, row in bulk_downloads_df.iterrows():
            table_id = row["table_id"]

            source_documentation_url = _guess_source_documentation_url(table_id)

            # Get description of the table
            # TODO - For now this is page scraping the description from the source_documentation_url page
            # In the future we should retrieve the description by finding a suitable API call.
            # The relevant API is here "https://www.nomisweb.co.uk/api/v01/"
            description = row["table_name"]
            description = _retrieve_table_description(source_documentation_url)

            # For now this does not use the "extra_post_release_filename" and "extra_post_release_url" tables
            for geo_level in self.geo_levels:
                # get the path within the zip file
                archive_file_path = _guess_csv_filename(
                    row["original_release_filename"], geo_level
                )

                catalog_summary["node"].append(bulk_downloads_index)
                catalog_summary["table_id"].append(table_id)
                catalog_summary["geo_level"].append(geo_level)
                catalog_summary["partition_key"].append(f"{geo_level}/{table_id}")
                catalog_summary["human_readable_name"].append(row["table_name"])
                catalog_summary["description"].append(description)
                catalog_summary["metric_parquet_file_url"].append(None)
                catalog_summary["parquet_column_name"].append(None)
                catalog_summary["parquet_margin_of_error_column"].append(None)
                catalog_summary["parquet_margin_of_error_file"].append(None)
                catalog_summary["potential_denominator_ids"].append(None)
                catalog_summary["parent_metric_id"].append(None)
                catalog_summary["source_data_release_id"].append(None)
                catalog_summary["source_download_url"].append(
                    row["original_release_url"]
                )
                catalog_summary["source_format"].append(None)
                catalog_summary["source_archive_file_path"].append(archive_file_path)
                catalog_summary["source_documentation_url"].append(
                    source_documentation_url
                )

        catalog_df = pd.DataFrame.from_records(catalog_summary)
        self.add_partition_keys(context, catalog_df["partition_key"].to_list())

        add_metadata(context, catalog_df, "Catalog")
        return catalog_df

    def _census_tables(self, context, catalog: pd.DataFrame) -> pd.DataFrame:
        """
        WIP:
        At present this function will download each zipfile multiple times. This is a consequence
        of the fact that
        * Each partition is a unique combination of topic summary and geometry level.
        * Each zip file contains multiple files including every geometry level for a given topic
          summary.

        I cannot see an easy way to share a cached downloaded zip file where different files are
        then extracted for different partitions (given dagster can in-theory execute partitions
        in arbitrary order and even on a cluster of machines). However a better solution may exist.
        """
        partition_key = context.asset_partition_key_for_output()
        current_table = catalog[catalog["partition_key"] == partition_key]

        source_download_url = current_table["source_download_url"].to_numpy()[0]
        source_archive_file_path = current_table["source_archive_file_path"].to_numpy()[
            0
        ]

        def _extract_csv_from_zipfile(temp_zip: Path, archive_path: Path) -> str:
            # This function works around certain bugs in the upstream source data
            #
            # 1) in some of the filenames have two consecutive `.` in the filename
            # e.g. `census2021-ts002-lsoa..csv`
            #
            # 2) In some cases the individual csv file exists within zipfile, but
            # is zero-sized. We treat this the same as if the file does not exist.
            #
            # 3) In at least one case the zipfile itself is malformed and cannot
            # be read.
            try:
                extract_file_path = extract_main_file_from_zip(
                    temp_zip, Path(temp_dir), archive_path
                )
            except ValueError:
                source_archive_file_path = str(Path(archive_path).with_suffix("..csv"))
                # If the file does not exist this second call will fail and we
                # allow the exception to propagate.
                extract_file_path = extract_main_file_from_zip(
                    temp_zip, Path(temp_dir), source_archive_file_path
                )
            except zipfile.BadZipFile as bzf:
                err_msg = "The downloaded zipfile is malformed and cannot be read."
                raise ValueError(err_msg) from bzf

            ic(extract_file_path)
            if Path(extract_file_path).stat().st_size == 0:
                err_msg = f"File {archive_path} exists but is zero-sized. Cannot read."
                raise ValueError(err_msg)

            return extract_file_path

        # Keep the temp directory when developing, for debugging purposes
        # del_temp_dir = os.getenv("ENV") != "dev"
        with TemporaryDirectory() as temp_dir:  # pyright: ignore [reportCallIssue]
            ic(source_download_url)
            temp_zip = Path(_download_zipfile(source_download_url, temp_dir))

            extract_file_path = None
            try:
                extract_file_path = _extract_csv_from_zipfile(
                    temp_zip, source_archive_file_path
                )
            except ValueError:
                # If we still can't find the file, then we assume that this combination of
                # table and geometry level does not exist.
                # We believe that there are ~63 cases (out of a possible 462) where this is the case
                # There does not seems to be a way to determine prior to this point in the
                # pipeline.
                # In this case we will return an empty dataframe
                err_msg = (
                    f"Unable to find the file `{source_archive_file_path}` in the"
                    f" zip file: {source_download_url}. Assuming this combination"
                    " of topic summary and geometry level does not exist."
                )
                context.log.warning(err_msg)
                # context.instance.delete_dynamic_partition(self.partition_name, partition_key)
                return pd.DataFrame()

            # If we still can't find the file, then there is a different, unforeseen problem
            if not extract_file_path:
                err_msg = f"Unable to find the file `{source_archive_file_path}` in the zip file: {source_download_url}"
                raise SourceDataAssumptionsOutdated(err_msg)

            census_table = pd.read_csv(extract_file_path)

        add_metadata(context, census_table, title=partition_key)
        return census_table

    def _derived_metrics(
        self, context, census_tables: DataFrame, source_metric_metadata: MetricMetadata
    ) -> MetricsOutput:
        _SEP = "_"
        partition_key = context.partition_key
        _geo_level = partition_key.split("/")[0]
        source_table = census_tables
        source_mmd = source_metric_metadata
        # source_column = source_mmd.parquet_column_name
        # context.log.debug(ic(source_table.columns))
        # context.log.debug(ic(source_column))
        # context.log.debug(ic(source_table.head()))
        context.log.debug(ic(len(source_table)))

        if len(source_table) == 0:
            # Source data not available
            msg = f"Source data not available for partition key: {partition_key}"
            context.log.warning(msg)
            return MetricsOutput(metadata=[], metrics=pd.DataFrame())

        # geo_id = EW_CENSUS_GEO_LEVELS[geo_level].census_table_column
        # The columns "date", "geography" and "geography code" appear to be common to all tables
        # TODO - Could we use the "date" column to update the mmds?
        geo_id = "geography code"
        source_table = source_table.rename(columns={geo_id: COL.GEO_ID.value}).drop(
            # Drop the "geography" column as it is the name, not the ID
            columns=["date", "geography"]
        )

        parquet_file_name = (
            f"{self.key_prefix}/metrics/"
            f"{''.join(c for c in partition_key if c.isalnum()) + '.parquet'}"
        )
        derived_mmd: list[MetricMetadata] = []

        # Create MMD for all of the existing columns
        for col in source_table.columns:
            # We do not need an MMD for the GEO_ID column
            if col == COL.GEO_ID.value:
                continue
            new_mmd = source_mmd.model_copy(deep=True)
            new_mmd.parent_metric_id = source_mmd.source_metric_id
            new_mmd.metric_parquet_path = parquet_file_name
            new_mmd.hxl_tag = "#population+details_unknown"
            new_mmd.parquet_column_name = str(col)
            new_mmd.human_readable_name = str(col)
            derived_mmd.append(new_mmd)

        # Now deal with the derived columns - create MMD and functions to calculate them
        new_column_funcs = {}

        # Filter function to sum the columns (which will be used in the loop below)
        def sum_cols_func(df, col_names: Iterable[str]):
            return df[col_names].sum(axis=1)
            #  sum([df[col] for col in col_names])

        try:
            metric_specs = DERIVED_COLUMN_SPECIFICATIONS[partition_key]
            for metric_spec in metric_specs:
                # Get the list of columns that need to be summed
                columns_to_sum = metric_spec.column_select(source_table)
                # ic(type(columns_to_sum))
                # ic(columns_to_sum)
                # Add details to the dict of new columns that will be created
                new_column_funcs[metric_spec.output_column_name] = partial(
                    sum_cols_func, col_names=columns_to_sum
                )

                # Create a new metric metadata object
                new_mmd = source_mmd.model_copy(deep=True)
                new_mmd.parent_metric_id = source_mmd.source_metric_id
                new_mmd.metric_parquet_path = parquet_file_name
                new_mmd.hxl_tag = metric_spec.hxltag
                new_mmd.parquet_column_name = metric_spec.output_column_name
                new_mmd.human_readable_name = metric_spec.human_readable_name
                derived_mmd.append(new_mmd)
        except KeyError:
            # No extra derived metrics specified for this partition -- only use
            # those from pivoted data
            pass

        # Create a new table which only has the GEO_ID and the new columns
        new_table = source_table.assign(**new_column_funcs)

        ic(len(derived_mmd))
        ic(len(new_table.columns))
        ic(type(context))

        # TODO - adding metadata does not work in the tests
        # checking the type of the `context` object is a workarroun for this
        # A more robust solution is required
        if not isinstance(
            context,
            dagster._core.execution.context.invocation.DirectAssetExecutionContext,
        ):
            context.add_output_metadata(
                metadata={
                    "metadata_preview": MetadataValue.md(
                        metadata_to_dataframe(derived_mmd).head().to_markdown()
                    ),
                    "metrics_shape": f"{new_table.shape[0]} rows x {new_table.shape[1]} columns",
                    "metrics_columns": MetadataValue.json(new_table.columns.to_list()),
                    "metrics_preview": MetadataValue.md(new_table.head().to_markdown()),
                },
            )

        return MetricsOutput(metadata=derived_mmd, metrics=new_table)

    def _source_data_releases(
        self,
        _context,
        geometry: list[GeometryOutput],
        data_publisher: DataPublisher,
    ) -> dict[str, SourceDataRelease]:
        source_data_releases = {}

        for geo_output in geometry:
            geo_metadata = geo_output.metadata
            source_data_release: SourceDataRelease = SourceDataRelease(
                name="Census 2021",
                date_published=date(2022, 6, 28),
                reference_period_start=CENSUS_COLLECTION_DATE,
                reference_period_end=CENSUS_COLLECTION_DATE,
                collection_period_start=CENSUS_COLLECTION_DATE,
                collection_period_end=CENSUS_COLLECTION_DATE,
                expect_next_update=date(2031, 1, 1),
                url="https://www.ons.gov.uk/census",
                data_publisher_id=data_publisher.id,
                # Taken from https://www.ons.gov.uk/census
                description="The census takes place every 10 years. It gives us a picture of all the people and households in England and Wales.",
                geometry_metadata_id=geo_metadata.id,
            )
            source_data_releases[geo_metadata.level] = source_data_release
        return source_data_releases

    def _source_metric_metadata(
        self,
        context,
        catalog: pd.DataFrame,
        source_data_releases: dict[str, SourceDataRelease],
    ) -> MetricMetadata:
        partition_key = context.partition_key
        if (
            self.required_tables is not None
            and partition_key not in DERIVED_COLUMN_SPECIFICATIONS
        ):
            skip_reason = (
                f"Skipping as requested partition {partition_key} is not configured "
                f"for derived metrics {DERIVED_COLUMN_SPECIFICATIONS.keys()}"
            )
            context.log.warning(skip_reason)
            raise RuntimeError(skip_reason)

        catalog_row = catalog[catalog["partition_key"] == partition_key].to_dict(
            orient="records"
        )[0]

        geo_level = catalog_row["geo_level"]
        source_table = SourceTable(
            # TODO: how programmatically do this
            hxltag="TBD",
            geo_level=geo_level,
            geo_column=EW_CENSUS_GEO_LEVELS[geo_level].geo_id_column,
            source_column="Count",
        )

        return census_table_metadata(
            catalog_row,
            source_table,
            source_data_releases,
        )

    def _geometry(self, context) -> list[GeometryOutput]:
        # TODO: This is almost identical to Northern Ireland and Belgium so can probably be refactored to common
        # function with config of releases and languages
        geometries_to_return = []
        for level_details in EW_CENSUS_GEO_LEVELS.values():
            # TODO: get correct values
            geometry_metadata = GeometryMetadata(
                validity_period_start=CENSUS_COLLECTION_DATE,
                validity_period_end=CENSUS_COLLECTION_DATE,
                level=level_details.level,
                hxl_tag=level_details.hxl_tag,
                country_metadata=self.country_metadata,
            )
            geometries_raw: gpd.GeoDataFrame = gpd.read_file(
                level_details.data_download_url
            )

            context.log.debug(ic(level_details))
            context.log.debug(ic(geometries_raw.head(1).T))

            # Standardised the column names
            geometries_gdf = geometries_raw.rename(
                columns={level_details.geo_id_column: "GEO_ID"}
            ).loc[:, ["geometry", "GEO_ID"]]
            name_lookup_df = (
                geometries_raw.rename(
                    columns={
                        level_details.geo_id_column: "GEO_ID",
                        level_details.name_columns["eng"]: "eng",
                    }
                )
                .loc[:, ["GEO_ID", "eng"]]
                .drop_duplicates()
            )
            geometries_to_return.append(
                GeometryOutput(
                    metadata=geometry_metadata,
                    gdf=geometries_gdf,
                    names_df=name_lookup_df,
                )
            )

        # Add output metadata
        # TODO, It is not clear that this is the best way to represent the metadata
        # Specifically, this assumes that the order of EW_CENSUS_GEO_LEVELS is based
        # on the hierarchy of the geometries, which may not be the case.
        example_geometry_output = geometries_to_return[0]
        first_metadata = example_geometry_output.metadata
        first_gdf = example_geometry_output.gdf
        first_names = example_geometry_output.names_df
        first_joined_gdf = first_gdf.merge(first_names, on="GEO_ID")
        ax = first_joined_gdf.plot(column="eng", legend=False)
        ax.set_title(f"England & Wales 2021 {first_metadata.level}")
        md_plot = markdown_from_plot()
        context.add_output_metadata(
            metadata={
                "all_geom_levels": MetadataValue.md(
                    ",".join(
                        [
                            geom_output.metadata.level
                            for geom_output in geometries_to_return
                        ]
                    )
                ),
                "first_geometry_plot": MetadataValue.md(md_plot),
                "first_names_preview": MetadataValue.md(
                    first_names.head().to_markdown()
                ),
            }
        )

        return geometries_to_return


def _guess_source_documentation_url(table_id):
    return f"https://www.nomisweb.co.uk/datasets/c2021{table_id.lower()}"


def _retrieve_table_description(source_documentation_url):
    soup = BeautifulSoup(
        requests.get(source_documentation_url).content, features="lxml"
    )
    landing_info = soup.find_all(id="dataset-landing-information")

    try:
        assert len(landing_info) == 1
        landing_info = landing_info[0]
    except AssertionError as ae:
        err_msg = f"Expected a single section with `id=dataset-landing-information`, but found {len(landing_info)}."
        raise SourceDataAssumptionsOutdated(err_msg) from ae

    return "\n".join([text.strip() for text in landing_info.stripped_strings])


def _guess_csv_filename(zip_filename, geometry_level):
    """
    Guess the name of the main file in the zip file.
    """
    stem = Path(zip_filename).stem
    return f"{stem}-{geometry_level}.csv"


def bulk_downloads_webpage() -> pd.DataFrame:
    """
    Get the list of bulk zip files from the bulk downloads page.
    """
    bulk_downloads_page = "https://www.nomisweb.co.uk/sources/census_2021_bulk"
    columns = ["table_id", "description", "original_release", "extra_post_release"]
    dfs = pd.read_html(bulk_downloads_page, header=0, extract_links="all")

    if len(dfs) != 1:
        err_msg = f"Expected a single table on the bulk downloads page, but found {len(dfs)} tables."
        raise SourceDataAssumptionsOutdated(err_msg)

    # The first table is the one we want
    download_df = dfs[0]
    download_df.columns = columns

    # There are some subheadings in the table, which are added as rows by `read_html`
    # These can be identified by the `table_id` == `description` == `original_release_filename`
    # We need to drop these rows
    download_df = download_df[download_df["table_id"] != download_df["description"]]
    # expand the tuples into individual columns
    return _expand_tuples_in_df(download_df)


def _expand_tuples_in_df(df) -> pd.DataFrame:
    """
    Expand the tuples in the DataFrame.
    """
    root_url = "https://www.nomisweb.co.uk/"

    columns = [
        "table_id",
        "table_name",
        "original_release_filename",
        "original_release_url",
        "extra_post_release_filename",
        "extra_post_release_url",
    ]
    new_df = pd.DataFrame(columns=columns)

    # Copy individual columns from the tuples
    # If there is a URL, it is in the second element of the tuple, and should be joined with the root URL
    # "table_id" and "description" do not have URLs
    new_df["table_id"] = df["table_id"].apply(lambda x: x[0])
    new_df["table_name"] = df["description"].apply(lambda x: x[0])
    new_df["original_release_filename"] = df["original_release"].apply(lambda x: x[0])
    new_df["original_release_url"] = df["original_release"].apply(
        lambda x: urljoin(root_url, x[1])
    )

    # There may not be a valid value for "extra_post_release", hence the check using `isinstance`
    new_df["extra_post_release_filename"] = df["extra_post_release"].apply(
        lambda x: x[0] if isinstance(x, tuple) else None
    )
    new_df["extra_post_release_url"] = df["extra_post_release"].apply(
        lambda x: urljoin(root_url, x[1]) if isinstance(x, tuple) else None
    )

    return new_df


def _download_zipfile(source_download_url, temp_dir) -> str:
    temp_dir = Path(temp_dir)
    temp_file = temp_dir / "data.zip"

    with requests.get(source_download_url, stream=True) as r:
        r.raise_for_status()
        with Path(temp_file).open(mode="wb") as f:
            for chunk in r.iter_content(chunk_size=(16 * 1024 * 1024)):
                f.write(chunk)

    return str(temp_file.resolve())


# if __name__ == "__main__":
#     # This is for testing only
#     # bulk_files_df = bulk_downloads_webpage()
#     # bulk_files_df = bulk_files_df.head(2)
#     # ic(bulk_files_df)

#     download_zip_files(bulk_files_df)


# Assets
ew_census = EnglandAndWales()
country_metadata = ew_census.create_country_metadata()
data_publisher = ew_census.create_data_publisher()
geometry = ew_census.create_geometry()
source_data_releases = ew_census.create_source_data_releases()
catalog = ew_census.create_catalog()
census_tables = ew_census.create_census_tables()
source_metric_metadata = ew_census.create_source_metric_metadata()
derived_metrics = ew_census.create_derived_metrics()
metrics = ew_census.create_metrics()



from __future__ import annotations

from dataclasses import dataclass

import geopandas as gpd
import pandas as pd
from dagster import AssetsDefinition

from poppusher.metadata import GeometryMetadata, MetricMetadata

from .sensor_class import CloudAssetSensor


@dataclass
class GeometryOutput:
    """This class conceptualises the expected output types of a geometry
    asset. Specifically, the asset marked with `@send_to_geometry_sensor` has to
    output a list of GeometryOutput objects (one per geometry level / year)."""

    metadata: GeometryMetadata
    gdf: gpd.GeoDataFrame
    names_df: pd.DataFrame


@dataclass
class MetricsOutput:
    """This class conceptualises the expected output types of a metrics
    asset. Specifically, the asset marked with `@send_to_metrics_sensor` has to
    output a list of MetricsOutput objects (one per parquet file; but each
    MetricsOutput object may correspond to multiple metrics that are serialised
    to the same parquet file)."""

    metadata: list[MetricMetadata]
    metrics: pd.DataFrame


metadata_factory = CloudAssetSensor(
    io_manager_key="metadata_io_manager",
    prefix="metadata",
    interval=20,
)

metadata_sensor = metadata_factory.create_sensor()
metadata_asset = metadata_factory.create_publishing_asset()

geometry_factory = CloudAssetSensor(
    io_manager_key="geometry_io_manager",
    prefix="geometry",
    interval=60,
)

geometry_sensor = geometry_factory.create_sensor()
geometry_asset = geometry_factory.create_publishing_asset()


metrics_factory = CloudAssetSensor(
    io_manager_key="metrics_io_manager",
    prefix="metrics",
    interval=60,
)

metrics_sensor = metrics_factory.create_sensor()
metrics_asset = metrics_factory.create_publishing_asset()

metrics_partitioned_factory = CloudAssetSensor(
    io_manager_key="metrics_partitioned_io_manager",
    prefix="metrics_partitioned",
    interval=60,
)

# TODO: commented out until implemented for partitioned assets
# metrics_partitioned_sensor = metrics_partitioned_factory.create_sensor()
# metrics_partitioned_asset = metrics_partitioned_factory.create_publishing_asset()

metrics_metadata_factory = CloudAssetSensor(
    io_manager_key="metrics_metadata_io_manager",
    prefix="metrics_metadata",
    interval=60,
)

metrics_metadata_sensor = metrics_metadata_factory.create_sensor()
metrics_metadata_asset = metrics_metadata_factory.create_publishing_asset()


def send_to_metadata_sensor(asset: AssetsDefinition):
    metadata_factory.monitored_asset_keys.append(asset.key)
    return asset


def send_to_geometry_sensor(asset: AssetsDefinition):
    geometry_factory.monitored_asset_keys.append(asset.key)
    return asset


def send_to_metrics_sensor(asset: AssetsDefinition):
    metrics_factory.monitored_asset_keys.append(asset.key)
    return asset


# TODO: need to implement handling for partitions for this sensor
def send_to_metrics_partitioned_sensor(asset: AssetsDefinition):
    # TODO: add partition key here
    metrics_partitioned_factory.monitored_asset_keys.append(asset.key)
    return asset


def send_to_metrics_metadata_sensor(asset: AssetsDefinition):
    metrics_metadata_factory.monitored_asset_keys.append(asset.key)
    return asset



from __future__ import annotations

import asyncio
import io
import os
from collections.abc import Callable
from dataclasses import dataclass
from datetime import date
from functools import reduce
from typing import ClassVar

import aiohttp
import geopandas as gpd
import pandas as pd
import requests
from bs4 import BeautifulSoup
from dagster import (
    MetadataValue,
)
from icecream import ic

from poppusher.assets.country import Country
from poppusher.cloud_outputs import GeometryOutput, MetricsOutput
from poppusher.metadata import (
    COL,
    CountryMetadata,
    DataPublisher,
    GeometryMetadata,
    MetricMetadata,
    SourceDataRelease,
    metadata_to_dataframe,
)
from poppusher.utils import add_metadata, markdown_from_plot


@dataclass
class NIGeometryLevel:
    level: str
    hxl_tag: str
    geo_id_column: str
    census_table_column: str
    name_columns: dict[str, str]  # keys = language codes, values = column names
    url: str
    lookup_url: str | None
    lookup_sheet: str | None
    left_on: str | None
    right_on: str | None


# Geometry levels to include
# Full list of geographies, see metadata:
# https://build.nisra.gov.uk/en/metadata/dataset?d=PEOPLE
# - "LGD14",  # Local Government District 2014
# - "URBAN_STATUS", # Urban Status
# - "HEALTH_TRUST", # Health and Social Care Trust
# - "PARLCON24", # Parliamentary Constituency 2024
# - "DEA14", # District Electoral Area 2014
# - "SDZ21",  # Census 2021 Super Data Zone
# - "DZ21",  # Census 2021 Data Zone
NI_GEO_LEVELS = {
    "DZ21": NIGeometryLevel(
        level="DZ21",
        hxl_tag="TBD",
        geo_id_column="DZ2021_cd",
        census_table_column="Census 2021 Data Zone Code",
        name_columns={"eng": "DZ2021_nm"},
        url="https://www.nisra.gov.uk/sites/nisra.gov.uk/files/publications/geography-dz2021-esri-shapefile.zip",
        lookup_url=None,
        lookup_sheet=None,
        left_on=None,
        right_on=None,
    ),
    "SDZ21": NIGeometryLevel(
        level="SDZ21",
        hxl_tag="TBD",
        geo_id_column="SDZ2021_cd",
        census_table_column="Census 2021 Super Data Zone Code",
        name_columns={"eng": "SDZ2021_nm"},
        url="https://www.nisra.gov.uk/sites/nisra.gov.uk/files/publications/geography-sdz2021-esri-shapefile.zip",
        lookup_url=None,
        lookup_sheet=None,
        left_on=None,
        right_on=None,
    ),
    "LGD14": NIGeometryLevel(
        level="LGD14",
        hxl_tag="TBD",
        geo_id_column="LGD2014_cd",
        census_table_column="Local Government District 2014 Code",
        name_columns={"eng": "LGD2014_name"},
        url="https://www.nisra.gov.uk/sites/nisra.gov.uk/files/publications/geography-dz2021-esri-shapefile.zip",
        lookup_url="https://www.nisra.gov.uk/sites/nisra.gov.uk/files/publications/geography-data-zone-and-super-data-zone-lookups.xlsx",
        lookup_sheet="DZ2021_lookup",
        left_on="DZ2021_cd",
        right_on="DZ2021_code",
    ),
}

# Required tables
TABLES_TO_PROCESS = ["MS-A09", "DT-0018"] if os.getenv("ENV") == "dev" else None

# 2021 census collection date
CENSUS_COLLECTION_DATE = date(2021, 3, 21)


async def get_nodes_and_links() -> dict[str, dict[str, str]]:
    """Extracts the URLs for census tables and metadata for ready-made tables."""
    SCHEME_AND_HOST = "https://build.nisra.gov.uk"
    urls = [
        "".join([SCHEME_AND_HOST, url.get("href")])
        for url in BeautifulSoup(
            requests.get(SCHEME_AND_HOST + "/en/standard").content, features="lxml"
        ).find_all("a")
        if str(url.get("href")).startswith("/en/standard")
    ]

    async def get_links(url: str, session: aiohttp.ClientSession) -> dict[str, str]:
        async with session.get(url) as response:
            soup = BeautifulSoup(await response.text(), features="lxml")
            return {
                "table_url": next(
                    iter(
                        [
                            "".join([SCHEME_AND_HOST, link.get("href")])
                            for link in soup.find_all("a")
                            if "table.csv?" in link.get("href")
                        ]
                    )
                ),
                "metadata_url": next(
                    iter(
                        [
                            "".join([SCHEME_AND_HOST, link.get("href")])
                            for link in soup.find_all("a")
                            if "table.csv-metadata" in link.get("href")
                        ]
                    )
                ),
            }

    async with aiohttp.ClientSession() as session, asyncio.TaskGroup() as tg:
        tasks = {url: tg.create_task(get_links(url, session)) for url in urls}
    return {url: task.result() for url, task in tasks.items()}


@dataclass
class DerivedColumn:
    hxltag: str
    filter_func: Callable[[pd.DataFrame], pd.DataFrame]
    output_column_name: str
    human_readable_name: str


@dataclass
class SourceTable:
    hxltag: str
    geo_level: str
    geo_column: str
    source_column: str


# Config for each partition to be derived
age_code = "`Age Code`"
sex_label = "`Sex Label`"
DERIVED_COLUMNS = [
    DerivedColumn(
        hxltag="#population+children+age5_17",
        filter_func=lambda df: df.query(f"{age_code} >= 5 and {age_code} < 18"),
        output_column_name="children_5_17",
        human_readable_name="Children aged 5 to 17",
    ),
    DerivedColumn(
        hxltag="#population+infants+age0_4",
        filter_func=lambda df: df.query(f"{age_code} >= 0 and {age_code} < 5"),
        output_column_name="infants_0_4",
        human_readable_name="Infants aged 0 to 4",
    ),
    DerivedColumn(
        hxltag="#population+children+age0_17",
        filter_func=lambda df: df.query(f"{age_code} >= 0 and {age_code} < 18"),
        output_column_name="children_0_17",
        human_readable_name="Children aged 0 to 17",
    ),
    DerivedColumn(
        hxltag="#population+adults+f",
        filter_func=lambda df: df.query(
            f"{age_code} >= 18 and {sex_label} == 'Female'"
        ),
        output_column_name="adults_f",
        human_readable_name="Female adults",
    ),
    DerivedColumn(
        hxltag="#population+adults+m",
        filter_func=lambda df: df.query(f"{age_code} >= 18 and {sex_label} == 'Male'"),
        output_column_name="adults_m",
        human_readable_name="Male adults",
    ),
    DerivedColumn(
        hxltag="#population+adults",
        filter_func=lambda df: df.query(f"{age_code} >= 18"),
        output_column_name="adults",
        human_readable_name="Adults",
    ),
    DerivedColumn(
        hxltag="#population+ind",
        filter_func=lambda df: df,
        output_column_name="individuals",
        human_readable_name="Total individuals",
    ),
]

DERIVED_COLUMN_SPECIFICATIONS: dict[str, list[DerivedColumn]] = {
    "DZ21/MS-A09": DERIVED_COLUMNS,
    "SDZ21/MS-A09": DERIVED_COLUMNS,
}


def census_table_metadata(
    catalog_row: dict[str, str],
    source_table: SourceTable,
    source_data_releases: dict[str, SourceDataRelease],
) -> MetricMetadata:
    return MetricMetadata(
        human_readable_name=catalog_row["human_readable_name"],
        source_download_url=catalog_row["source_download_url"],
        source_archive_file_path=catalog_row["source_archive_file_path"],
        source_documentation_url=catalog_row["source_documentation_url"],
        source_data_release_id=source_data_releases[source_table.geo_level].id,
        # TODO - this is a placeholder
        parent_metric_id="unknown_at_this_stage",
        potential_denominator_ids=None,
        parquet_margin_of_error_file=None,
        parquet_margin_of_error_column=None,
        parquet_column_name=source_table.source_column,
        # TODO - this is a placeholder
        metric_parquet_path="unknown_at_this_stage",
        hxl_tag=source_table.hxltag,
        description=catalog_row["description"],
        source_metric_id=source_table.hxltag,
    )


class NorthernIreland(Country):
    country_metadata: CountryMetadata = CountryMetadata(
        name_short_en="Northern Ireland",
        name_official="Northern Ireland",
        iso3="GBR",
        iso2="GB",
        iso3166_2="GB-NIR",
    )
    geo_levels: ClassVar[list[str]] = list(NI_GEO_LEVELS.keys())
    tables_to_process: list[str] | None = TABLES_TO_PROCESS

    def _data_publisher(
        self, _context, country_metadata: CountryMetadata
    ) -> DataPublisher:
        return DataPublisher(
            name="NISRA",
            url="https://www.nisra.gov.uk/",
            description=(
                "The Northern Ireland Statistics and Research Agency (NISRA), which "
                "incorporates the General Register Office (GRO), is an executive agency "
                "within the Department of Finance (NI) and was established on 1 April 1996."
            ),
            countries_of_interest=[country_metadata.id],
        )

    def _catalog(self, context) -> pd.DataFrame:
        return asyncio.run(self._catalog_async(context))

    async def _catalog_async(self, context) -> pd.DataFrame:
        """
        A catalog for NI can be generated in two ways:
        1. With flexible table builder:
                https://build.nisra.gov.uk/en/
            with metadata chosen from:
                https://build.nisra.gov.uk/en/metadata
        2. Or through enumerating the ready-made tables:
            https://build.nisra.gov.uk/en/standard
            However, for some geographical resolutions, ready-made tables may
            not be available due to data confidentiality.
        """
        catalog_summary = {
            "node": [],
            "partition_key": [],
            "table_id": [],
            "geo_level": [],
            "human_readable_name": [],
            "description": [],
            "metric_parquet_file_url": [],
            "parquet_column_name": [],
            "parquet_margin_of_error_column": [],
            "parquet_margin_of_error_file": [],
            "potential_denominator_ids": [],
            "parent_metric_id": [],
            "source_data_release_id": [],
            "source_download_url": [],
            "source_format": [],
            "source_archive_file_path": [],
            "source_documentation_url": [],
            "table_schema": [],
        }
        nodes = await get_nodes_and_links()
        self.remove_all_partition_keys(context)

        def add_resolution(s: str, geo_level: str) -> str:
            s_split = s.split("?")
            query_params = s_split[1].split("&")
            if query_params[0].startswith("d="):
                query_params = "&".join(
                    [query_params[0], f"v={geo_level}", *query_params[2:]]
                )
            else:
                query_params = "&".join([f"v={geo_level}", *query_params[1:]])
            out_url = "?".join([s_split[0], query_params])
            ic(out_url)
            return out_url

        async def get_catalog_summary(
            node_url: str,
            node_items: dict[str, str],
            geo_level: str,
            session: aiohttp.ClientSession,
        ):
            async with session.get(node_items["metadata_url"]) as response:
                metadata = await response.json()
                table_id = metadata["dc:title"].split(":")[0]
                # Skip if not required
                if (
                    self.tables_to_process is not None
                    and table_id not in self.tables_to_process
                ):
                    return None
                return {
                    "node": node_url,
                    "table_id": table_id,
                    "geo_level": geo_level,
                    "partition_key": f"{geo_level}/{table_id}",
                    "human_readable_name": metadata["dc:title"],
                    "description": metadata["dc:description"],
                    "metric_parquet_file_url": None,
                    "parquet_column_name": None,
                    "parquet_margin_of_error_column": None,
                    "parquet_margin_of_error_file": None,
                    "potential_denominator_ids": None,
                    "parent_metric_id": None,
                    "source_data_release_id": None,
                    "source_download_url": add_resolution(metadata["url"], geo_level),
                    "source_format": None,
                    "source_archive_file_path": None,
                    "source_documentation_url": node_url,
                    "table_schema": metadata["tableSchema"],
                }

        async with aiohttp.ClientSession() as session, asyncio.TaskGroup() as tg:
            tasks = [
                tg.create_task(
                    get_catalog_summary(node_url, node_items, geo_level, session)
                )
                for node_url, node_items in nodes.items()
                for geo_level in self.geo_levels
            ]
        catalog_summary = [task.result() for task in tasks]
        # Remove the empty results for partitions that weren't required
        catalog_summary = [v for v in catalog_summary if v is not None]
        catalog_df = pd.DataFrame.from_records(catalog_summary)
        self.add_partition_keys(context, catalog_df["partition_key"].to_list())

        add_metadata(context, catalog_df, "Catalog")
        return catalog_df

    def _census_tables(self, context, catalog: pd.DataFrame) -> pd.DataFrame:
        partition = context.asset_partition_key_for_output()
        ic(partition)
        ic(catalog.loc[catalog["partition_key"].eq(partition), "source_download_url"])
        url = catalog.loc[
            catalog["partition_key"].eq(partition), "source_download_url"
        ].iloc[0]
        census_table = pd.read_csv(
            io.BytesIO(requests.get(url).content), encoding="utf8"
        )
        add_metadata(
            context, census_table, title=context.asset_partition_key_for_output()
        )
        return census_table

    def _geometry(self, context) -> list[GeometryOutput]:
        # TODO: This is almost identical to Belgium so can probably be refactored to common
        # function with config of releases and languages
        geometries_to_return = []
        for level_details in NI_GEO_LEVELS.values():
            # TODO: get correct values
            geometry_metadata = GeometryMetadata(
                country_metadata=self.country_metadata,
                validity_period_start=CENSUS_COLLECTION_DATE,
                validity_period_end=CENSUS_COLLECTION_DATE,
                level=level_details.level,
                hxl_tag=level_details.hxl_tag,
            )
            region_geometries_raw: gpd.GeoDataFrame = gpd.read_file(level_details.url)
            if level_details.lookup_url is not None:
                lookup = pd.read_excel(
                    level_details.lookup_url, sheet_name=level_details.lookup_sheet
                )
                region_geometries_raw = region_geometries_raw.merge(
                    lookup,
                    left_on=level_details.left_on,
                    right_on=level_details.right_on,
                    how="outer",
                )

            region_geometries_raw = region_geometries_raw.dissolve(
                by=level_details.geo_id_column
            ).reset_index()

            context.log.debug(ic(region_geometries_raw.head()))
            region_geometries = region_geometries_raw.rename(
                columns={level_details.geo_id_column: COL.GEO_ID.value}
            ).loc[:, ["geometry", COL.GEO_ID.value]]
            region_names = (
                region_geometries_raw.rename(
                    columns={
                        level_details.geo_id_column: COL.GEO_ID.value,
                        level_details.name_columns["eng"]: "eng",
                    }
                )
                .loc[:, [COL.GEO_ID.value, "eng"]]
                .drop_duplicates()
            )
            geometries_to_return.append(
                GeometryOutput(
                    metadata=geometry_metadata,
                    gdf=region_geometries,
                    names_df=region_names,
                )
            )

        # Add output metadata
        first_geometry = geometries_to_return[0]
        first_joined_gdf = first_geometry.gdf.merge(
            first_geometry.names_df, on=COL.GEO_ID.value
        )
        ax = first_joined_gdf.plot(column="eng", legend=False)
        ax.set_title(f"NI 2021 {first_geometry.metadata.level}")
        md_plot = markdown_from_plot()
        context.add_output_metadata(
            metadata={
                "all_geom_levels": MetadataValue.md(
                    ",".join(
                        [
                            geo_output.metadata.level
                            for geo_output in geometries_to_return
                        ]
                    )
                ),
                "first_geometry_plot": MetadataValue.md(md_plot),
                "first_names_preview": MetadataValue.md(
                    first_geometry.names_df.head().to_markdown()
                ),
            }
        )

        return geometries_to_return

    def _source_data_releases(
        self, _context, geometry, data_publisher
    ) -> dict[str, SourceDataRelease]:
        source_data_releases = {}
        for geom in geometry:
            source_data_release: SourceDataRelease = SourceDataRelease(
                name="Census 2021",
                # https://www.nisra.gov.uk/publications/census-2021-outputs-prospectus:
                # 9.30 am on 21 February 2023 for DZ and SDZ and District Electoral Areas
                date_published=date(2023, 2, 21),
                reference_period_start=date(2021, 3, 21),
                reference_period_end=date(2021, 3, 21),
                collection_period_start=CENSUS_COLLECTION_DATE,
                collection_period_end=CENSUS_COLLECTION_DATE,
                expect_next_update=date(2031, 1, 1),
                url="https://www.nisra.gov.uk/publications/census-2021-outputs-prospectus",
                data_publisher_id=data_publisher.id,
                description="TBC",
                geometry_metadata_id=geom.metadata.id,
            )
            source_data_releases[geom.metadata.level] = source_data_release
        return source_data_releases

    def _source_metric_metadata(
        self,
        context,
        catalog: pd.DataFrame,
        source_data_releases: dict[str, SourceDataRelease],
    ) -> MetricMetadata:
        partition_key = context.partition_key
        catalog_row = catalog[catalog["partition_key"] == partition_key].to_dict(
            orient="records"
        )[0]

        geo_level = partition_key.split("/")[0]
        source_table = SourceTable(
            # TODO: how programmatically do this
            hxltag="TBD",
            geo_level=geo_level,
            geo_column=NI_GEO_LEVELS[geo_level].geo_id_column,
            source_column="Count",
        )

        return census_table_metadata(
            catalog_row,
            source_table,
            source_data_releases,
        )

    def _derived_metrics(
        self,
        context,
        census_tables: pd.DataFrame,
        source_metric_metadata: MetricMetadata,
    ) -> MetricsOutput:
        SEP = "_"
        partition_key = context.partition_key
        geo_level = partition_key.split("/")[0]
        source_table = census_tables
        source_mmd = source_metric_metadata
        source_column = source_mmd.parquet_column_name

        if source_column not in source_table.columns or len(source_table) == 0:
            # Source data not available
            msg = f"Source data not available for partition key: {partition_key}"
            context.log.warning(msg)
            return MetricsOutput(metadata=[], metrics=pd.DataFrame())

        geo_id = NI_GEO_LEVELS[geo_level].census_table_column
        source_table = source_table.rename(columns={geo_id: COL.GEO_ID.value}).drop(
            columns=geo_id.replace("Code", "Label")
        )

        parquet_file_name = (
            f"{self.key_prefix}/metrics/"
            f"{''.join(c for c in partition_key if c.isalnum()) + '.parquet'}"
        )
        derived_metrics: list[pd.DataFrame] = []
        derived_mmd: list[MetricMetadata] = []

        try:
            metric_specs = DERIVED_COLUMN_SPECIFICATIONS[partition_key]
            for metric_spec in metric_specs:
                new_table = (
                    source_table.pipe(metric_spec.filter_func)
                    .groupby(by=COL.GEO_ID.value, as_index=True)
                    .sum()
                    .rename(columns={source_column: metric_spec.output_column_name})
                    .filter(items=[COL.GEO_ID.value, metric_spec.output_column_name])
                )
                derived_metrics.append(new_table)
                new_mmd = source_mmd.copy()
                new_mmd.parent_metric_id = source_mmd.source_metric_id
                new_mmd.metric_parquet_path = parquet_file_name
                new_mmd.hxl_tag = metric_spec.hxltag
                new_mmd.parquet_column_name = metric_spec.output_column_name
                new_mmd.human_readable_name = metric_spec.human_readable_name
                derived_mmd.append(new_mmd)
        except KeyError:
            # No extra derived metrics specified for this partition -- only use
            # those from pivoted data
            pass

        # Get all other metrics from table as is pivoted
        def pivot_df(df: pd.DataFrame, end: str) -> tuple[list[str], pd.DataFrame]:
            # Variables are either code or label, only keep the case for given 'end'
            cols = (
                [col for col in df.columns if col.endswith(end)]
                + [COL.GEO_ID.value]
                + ["Count"]
            )
            pivot_cols = [col for col in cols if col not in [COL.GEO_ID.value, "Count"]]
            ic(cols)
            ic(pivot_cols)
            ic(df.columns)
            ic(df.head())
            pivot = df[cols].pivot_table(
                index=COL.GEO_ID.value,
                columns=pivot_cols,
                values="Count",
                aggfunc="sum",
            )

            # FLattent multi-index
            if isinstance(pivot.columns, pd.MultiIndex):
                pivot.columns = [
                    SEP.join(list(map(str, col))).strip()
                    for col in pivot.columns.to_numpy()
                ]
            # Ensure columns are string
            else:
                pivot.columns = [str(col).strip() for col in pivot.columns.to_numpy()]
            out_cols = [col.replace(end, "").strip() for col in pivot_cols]
            return out_cols, pivot

        # Pivot for codes and labels
        for var_type in ["Code", "Label"]:
            out_cols, new_table = pivot_df(source_table, var_type)
            ic(new_table)
            for metric_col in new_table.columns:
                metric_df = new_table.loc[:, metric_col].to_frame()
                ic(metric_df)
                derived_metrics.append(metric_df)
                new_mmd = source_mmd.copy()
                new_mmd.parent_metric_id = source_mmd.source_metric_id
                new_mmd.metric_parquet_path = parquet_file_name
                # TODO: check this
                key_val = dict(zip(out_cols, metric_col.split(SEP), strict=True))

                def gen_hxltag(kv: dict[str, str]) -> str:
                    out = ["#population"]
                    for key, value in kv.items():
                        out += [
                            "".join(c for c in key if c.isalnum())
                            + "_"
                            + "".join(c for c in value if c.isalnum())
                        ]
                    return "+".join(out)

                new_mmd.hxl_tag = gen_hxltag(key_val)
                new_mmd.parquet_column_name = metric_col
                new_mmd.human_readable_name = "; ".join(
                    [
                        f"Variable: '{key}'; Value: '{value}'"
                        for key, value in key_val.items()
                    ]
                )
                derived_mmd.append(new_mmd)

        joined_metrics = reduce(
            lambda left, right: left.merge(
                right, on=COL.GEO_ID.value, how="inner", validate="one_to_one"
            ),
            derived_metrics,
        )

        context.add_output_metadata(
            metadata={
                "metadata_preview": MetadataValue.md(
                    metadata_to_dataframe(derived_mmd).head().to_markdown()
                ),
                "metrics_shape": f"{joined_metrics.shape[0]} rows x {joined_metrics.shape[1]} columns",
                "metrics_preview": MetadataValue.md(
                    joined_metrics.head().to_markdown()
                ),
            },
        )
        return MetricsOutput(metadata=derived_mmd, metrics=joined_metrics)


# Assets
ni = NorthernIreland()
country_metadata = ni.create_country_metadata()
data_publisher = ni.create_data_publisher()
geometry = ni.create_geometry()
source_data_releases = ni.create_source_data_releases()
catalog = ni.create_catalog()
census_tables = ni.create_census_tables()
source_metric_metadata = ni.create_source_metric_metadata()
derived_metrics = ni.create_derived_metrics()
metrics = ni.create_metrics()



import pandas as pd 
import urllib.request
import tempfile
import os 
from functools import reduce
from tqdm import tqdm
from more_itertools import batched
import geopandas as gp
import subprocess
import docker
from pathlib import Path
import os
import tempfile

SUMMARY_LEVELS={
    "tract": 140 ,
    "block_group":150,
    "county":50
}

ACS_METADATA={
    # 2018 needs additional logic for joining together geometry files
    # 2018:{
    #     base:"https://www2.census.gov/programs-surveys/acs/summary_file/2018",
    #     type:'table',
    #     geoms:{
    #         tract:,
    #         block:"",
    #         county:"https://www2.census.gov/geo/tiger/GENZ2018/shp/cb_2018_us_county_500k.zip":
    #     }
    #     oneYear:{
    #         tables:"prototype/1YRData/",
    #         geoIds:"https://www2.census.gov/programs-surveys/acs/summary_file/2018/prototype/20181YRGeos.csv",
    #         shells:"prototype/ACS2018_Table_Shells.xlsx"
    #     },
    #     fiveYear:{
    #         tables:"prototype/5YRData/",
    #         geoIds:"https://www2.census.gov/programs-surveys/acs/summary_file/2018/prototype/20185YRGeos.csv",
    #         shells:"prototype/ACS2018_Table_Shells.xlsx"
    #     }
    # },
    2019:{
        "base" : "https://www2.census.gov/programs-surveys/acs/summary_file/2019/",
        "type":'table',
        "geoms":{
            "tract": "https://www2.census.gov/geo/tiger/GENZ2019/shp/cb_2019_us_tract_500k.zip",
            "block_group":"https://www2.census.gov/geo/tiger/GENZ2019/shp/cb_2019_us_bg_500k.zip",
            "county": "https://www2.census.gov/geo/tiger/GENZ2019/shp/cb_2019_us_county_500k.zip"
        },
        "oneYear":{
            "tables":"prototype/1YRData/",
            "geoIds":"prototype/Geos20191YR.csv",
            "shells":"prototype/ACS2019_Table_Shells.csv"
        },
        "fiveYear":{
            "tables":"prototype/5YRData/",
            "geoIds":"prototype/Geos20195YR.csv",
            "shells":"prototype/ACS2019_Table_Shells.csv"
        },
        "geoIdCol": "DADSID"
    },
    # Note 1 year esimates are not avaliable because of covid. More details of the exeprimental estimates 
    # are here : https://www.census.gov/programs-surveys/acs/technical-documentation/table-and-geography-changes/2020/1-year.html
    2020:{
        "base":"https://www2.census.gov/programs-surveys/acs/summary_file/2020/",
        "type":'table',
        "geoms":{
            "tract":"https://www2.census.gov/geo/tiger/GENZ2020/shp/cb_2020_us_tract_500k.zip",
            "block_group":"https://www2.census.gov/geo/tiger/GENZ2020/shp/cb_2020_us_bg_500k.zip",
            "county":"https://www2.census.gov/geo/tiger/GENZ2020/shp/cb_2020_us_county_500k.zip"
        },
        "shells":"prototype/ACS2020_Table_Shells.csv",
        "oneYear":None,
        "fiveYear":{
            "shells":"prototype/ACS2020_Table_Shells.csv",
            "tables":"prototype/5YRData/",
            "geoIds":"prototype/Geos20205YR.csv",
        },
        "geoIdCol": "DADSID"
    },
    2021:{
        "base":"https://www2.census.gov/programs-surveys/acs/summary_file/2021/table-based-SF/",
        "type":'table',
        "geoIdsSep":"|",
        "geoms":{
            "tract":"https://www2.census.gov/geo/tiger/GENZ2021/shp/cb_2021_us_tract_500k.zip",
            "block_group":"https://www2.census.gov/geo/tiger/GENZ2021/shp/cb_2021_us_bg_500k.zip",
            "county":"https://www2.census.gov/geo/tiger/GENZ2021/shp/cb_2021_us_county_500k.zip"
        },
        "oneYear":{
            "tables":"data/1YRData/",
            "geoIds":"documentation/Geos20211YR.txt",
            "shells":"documentation/ACS20211YR_Table_Shells.txt"
        },
        "fiveYear":{
            "tables":"data/5YRData/",
            "geoIds":"documentation/Geos20215YR.txt",
            "shells":"documentation/ACS20215YR_Table_Shells.txt"
        },
        "geoIdCol": "GEO_ID"
    }
}

def generate_variable_dictionary(year:int, summary_level:str):
    metadata = ACS_METADATA[year]
    base = metadata["base"]
    shells = metadata[summary_level]["shells"]

    # Config for each year
    if int(year) == 2019:
        raw= pd.read_csv(base+shells, encoding="latin")
        unique_id_col_name = "UniqueID"
    elif int(year) == 2020:
        raw= pd.read_csv(base+shells, encoding="latin")
        unique_id_col_name = "Unique ID"
    elif int(year) == 2021:
        raw= pd.read_csv(base+shells, sep="|")
        unique_id_col_name = "Unique ID"
        raw = raw.rename(columns={"Label": "Stub"})
    else:
        raise ValueError(f"generate_variable_dictionary() not implemented for year: {year}")

    result = [] # pd.DataFrame(columns=["tableID",unique_id_col_name, "universe","tableName", "variableName", "variableExtedndedName"])
    universe =""
    tableName = ""
    path =[]
    previousWasEdge = True 
    for (index,row) in raw.iterrows():
        if(( type(row["Table ID"]) == str and len(row["Table ID"].strip())==0) or type(row["Table ID"]) == float):
            # path=[]
            # previousWasEdge = True 
            continue

        stub = row["Stub"]
        # Get table name
        if year != 2021:
            if row[[unique_id_col_name]].isna().all() and not row[["Data Release"]].isna().all():
                table_name = row["Stub"].title()
            elif row[[unique_id_col_name]].isna().all():
                if("Universe" in stub):
                    universe = stub.split("Universe:")[1].strip()
                else:
                    universe=stub
        # Get universe
        if not (row[[unique_id_col_name]].isna().all()):
            # Universe is a column for 2021
            if year == 2021:
                universe = row["Universe"]
                table_name = row["Title"].title()
            if (":" in stub):
                if(previousWasEdge):
                    path.append(stub.replace(":",""))
                else:
                    path.pop()
                    path.append(stub.replace(":",""))
            else:
                previousWasEdge = False 
            extendedName = "|".join(path) 
            if(":" not in stub):
                extendedName = extendedName + "|"+stub
            result.append({
                "tableID": row["Table ID"],
                "uniqueID":row[unique_id_col_name],
                "universe":universe,
                "tableName": table_name,
                "variableName":stub,
                "variableExtendedName": extendedName
            })
    
                
                

    return pd.DataFrame.from_records(result)


def download_cartography_file(year: int, admin_level: str, work_dir: str| None = None):
    metadata = ACS_METADATA[year]
    url = metadata['geoms'][admin_level]
    if(work_dir == None):
        work_dir = tempfile.mkdtemp()
    local_dir = os.path.join(work_dir, admin_level+".zip")
    urllib.request.urlretrieve(url, local_dir)
    return local_dir

def convert_cartography_file_to_formats(path: str):
    data=  gp.read_file(f"zip://{path}")
    data.to_parquet(path.replace(".zip",".parquet"))
    data.to_file(path.replace(".zip",".flatgeobuff"), driver="FlatGeobuf")
    data.to_file(path.replace(".zip", ".geojsonseq"), driver="GeoJSONSeq")

def generate_pmtiles(path:str):
    client = docker.from_env()
    mount_folder = Path(path).resolve()
    container =client.containers.run("stuartlynn/tippecanoe:latest",
                          "tippecanoe -o tracts.pmtiles tracts.geojsonseq",
                          volumes={mount_folder: {"bind":"/app","mode":"rw"} },
                          detach=True,
                          remove=True)

    output = container.attach(stdout=True, stream=True, logs=True)
    for line in output:
        print(line)

""""
    Return the fips codes for states as two digit zero padded strings.
    Exclude state id which are placeholders for future potential states. 
"""
def state_fips():
    return [ f'{n:02d}' for n in  range(1,56) if n not in [3,7,14,43,52]]

"""
    Convert a data table to parquet and output 
"""
def convert_to_parquet(files: list[str], metrics: [list[str]]): # type: ignore
    pass
    # variableDescs = get_variable_definitions()
    # for metric in metrics:
    #      print(variableDescs[metric])
    
    # for file in files:
    #     data = pd.read_csv(file, dtype={"GEO_ID":str})
    #     data.drop(["state","county","tract"],axis=1).to_parquet(file.replace(".csv",".parquet"))

"""
    get the names of each summary file for a given year and summary level
"""
def get_summary_table_file_names(year:int, summary_level:str="fiveYear"):
    metadata = ACS_METADATA[year]
    base = metadata['base']
    table_path =  base + metadata[summary_level]['tables'] 
    
    table = pd.read_html(table_path)[0]
    filtered = table[table['Name'].str.startswith("acs",na=False)]
    return list(filtered["Name"])

"""
    Get the geometry identifier table for a given year and summary_level
"""
def get_geom_ids_table_for_summary(year:int, summary_level:str):
    path = ACS_METADATA[year]["base"] + ACS_METADATA[year][summary_level]['geoIds']
    sep = ACS_METADATA[year]["geoIdsSep"] if "geoIdsSep" in ACS_METADATA[year] else ","
    table = pd.read_csv(path, encoding='latin', sep=sep, low_memory=False)
    return table

"""
    Extract variables for the geographies we are interested in from the summary table
"""
def extract_values_at_specified_levels(
        df: pd.DataFrame, geoids: pd.DataFrame, geo_ids_col: str = "DADSID"
    ):
    joined = pd.merge(df,geoids[[geo_ids_col,"SUMLEVEL"]], left_on="GEO_ID", right_on=geo_ids_col, how='left')
    result = {}

    for (level, id) in SUMMARY_LEVELS.items():
        result[level]=(
            joined[joined['SUMLEVEL']==id]
            .drop(
                ["SUMLEVEL"]
                + ([geo_ids_col] if geo_ids_col != "GEO_ID" else []), axis=1)
        )
    return result         
        

"""
    Download and process a specific summary table 
"""
def get_summary_table(table_name: str, year:int, summary_level:str):
    base = ACS_METADATA[year]["base"]
    summary_file_dir = base + ACS_METADATA[year][summary_level]['tables']
    data = pd.read_csv(f"{summary_file_dir}/{table_name}", sep="|")
    return data

def select_estimates(df):
    estimate_columns = [ col for col in df.columns if col[0]=="B" and col.split("_")[1][0]=="E"]
    return df[estimate_columns]

def select_errors(df):
    error_columns = [ col for col in df.columns if col[0]=="B" and col.split("_")[1][0]=="M"]
    return df[error_columns]


def merge_parquet_files(file_names):
    result=pd.DataFrame()
    for batch in tqdm(batched(file_names,20)):
        newDFS = [select_estimates(pd.read_parquet(file).set_index("GEO_ID")) for file in batch]
        result = pd.concat([result] + newDFS  ,axis=1)
    return result

"""
    Generate the metirc parquet files for a given year and summary level 
"""
def process_year_summary_level(year:int, summary_level:str ="fiveYear"):
    workdir = tempfile.mkdtemp()
    tractDir = os.path.join(workdir,"tract")
    blockGroupDir = os.path.join(workdir,"block_groups")
    countyDir = os.path.join(workdir,"counties")
    os.mkdir(blockGroupDir)
    os.mkdir(tractDir)
    os.mkdir(countyDir)

    
    table_names = get_summary_table_file_names(year,summary_level)
    geoids = get_geom_ids_table_for_summary(year,summary_level)
    print("temp dir is ", workdir)
    for table in tqdm(table_names):
        df = get_summary_table(table, year,summary_level)
        values = extract_values_at_specified_levels(df, geoids)
        values['tract'].to_parquet(os.path.join(tractDir,table.replace(".dat",".parquet")))
        values['county'].to_parquet(os.path.join(countyDir,table.replace(".dat",".parquet")))
        values['block_group'].to_parquet(os.path.join(blockGroupDir,table.replace(".dat",".parquet")))

    merge_parquet_files([os.path.join(countyDir,file) for file in os.listdir(countyDir)]).to_parquet(f"county_{year}_{summary_level}.parquet")
    merge_parquet_files([os.path.join(tractDir,file) for file in os.listdir(tractDir)]).to_parquet(f"tract_{year}_{summary_level}.parquet")
    merge_parquet_files([os.path.join(blockGroupDir,file) for file in os.listdir(blockGroupDir)]).to_parquet(f"block_groups_{year}_{summary_level}.parquet")

if __name__ == "__main__":
    process_year_summary_level(2019,"fiveYear")
# """
#     Old code for getting the metrics from the Census API
# """
# def get_census_merics(years:list[int], admin_level:str, metrics:list[str]):
#     import requests
#     import json
#     import csv

#     metrics = get_variable_definitions();
#     metrics = [k for k in metrics.keys() if k[0]=='D']
#     print(metrics)
    
#     for year in years:
#         allValues =[]
#         for state in state_fips():
#             print(f"Downloading {year} for {state}")
#             api_call = f'https://api.census.gov/data/{year}/acs/acs5/profile?get=GEO_ID,{",".join(metrics)}&for={admin_level}:*&in=state:{state}&in=county:*'

#             print(api_call)
#             try:
#                 r = requests.get(api_call)
#                 r.raise_for_status

#                 data = r.json()
#                 if(len(allValues) > 0):
#                     allValues = allValues + data[1:]
#                 else:
#                     allValues = data
#             except requests.exceptions.HTTPError as errh:
#                 print(errh)
                           

#         with open(f'acs5_{year}_by_tract.csv','w',newline='') as csvfile:
#             csv.writer(csvfile).writerows(allValues)
#     print("Got metric")



from __future__ import annotations

import os
import tempfile
from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path

import geopandas as gpd
import pandas as pd
from azure.core.credentials import (
    AzureSasCredential,
)
from azure.storage.filedatalake import (
    DataLakeLeaseClient,
    DataLakeServiceClient,
    FileSystemClient,
)
from dagster import (
    Any,
    InputContext,
    IOManager,
    OutputContext,
)
from dagster_azure.adls2.utils import ResourceNotFoundError, create_adls2_client
from dagster_azure.blob.utils import create_blob_client
from icecream import ic
from upath import UPath

from . import (
    GeoIOManager,
    MetadataIOManager,
    MetricsIOManager,
    MetricsMetdataIOManager,
    MetricsPartitionedIOManager,
)

# Set no time limit on lease duration to enable large files to be uploaded
_LEASE_DURATION = -1

# Set connection timeout to be larger than default:
# https://github.com/Azure/azure-sdk-for-python/issues/26993#issuecomment-1289799860
_CONNECTION_TIMEOUT = 6000


class AzureMixin:
    storage_account: str | None = os.getenv("AZURE_STORAGE_ACCOUNT")
    container: str | None = os.getenv("AZURE_CONTAINER")
    prefix: str | None = os.getenv("AZURE_DIRECTORY")
    sas_token: str | None = os.getenv("SAS_TOKEN")
    adls2_client: DataLakeServiceClient
    file_system_client: FileSystemClient

    def __init__(self):
        if self.storage_account is None:
            err_msg = "Storage account needs to be provided."
            raise ValueError(err_msg)
        if self.sas_token is None:
            err_msg = "Credenital (SAS) needs to be provided."
            raise ValueError(err_msg)
        if self.container is None:
            err_msg = "Container needs to be provided."
            raise ValueError(err_msg)

        self.adls2_client = create_adls2_client(
            self.storage_account, AzureSasCredential(self.sas_token)
        )
        self.file_system_client = self.adls2_client.get_file_system_client(
            self.container
        )
        # Blob client needed to handle copying as ADLS doesn't have a copy API yet
        self.blob_client = create_blob_client(
            self.storage_account, AzureSasCredential(self.sas_token)
        )
        self.blob_container_client = self.blob_client.get_container_client(
            self.container
        )

        self.lease_duration = _LEASE_DURATION
        self.file_system_client.get_file_system_properties()

    def get_base_path(self) -> UPath:
        return UPath(self.prefix)

    @property
    def lease_client_constructor(self) -> Any:
        return DataLakeLeaseClient

    def _uri_for_path(self, path: UPath, protocol: str = "abfss://") -> str:
        return "{protocol}{filesystem}@{account}.dfs.core.windows.net/{key}".format(
            protocol=protocol,
            filesystem=self.file_system_client.file_system_name,
            account=self.file_system_client.account_name,
            key=path.as_posix(),
        )

    def get_loading_input_log_message(self, path: UPath) -> str:
        return f"Loading ADLS2 object from: {self._uri_for_path(path)}"

    def get_writing_output_log_message(self, path: UPath) -> str:
        return f"Writing ADLS2 object at: {self._uri_for_path(path)}"

    def unlink(self, path: UPath) -> None:
        file_client = self.file_system_client.get_file_client(path.as_posix())
        with self._acquire_lease(file_client, is_rm=True) as lease:
            file_client.delete_file(lease=lease, recursive=True)

    def dump_to_path(self, context: OutputContext, obj: bytes, path: UPath) -> None:
        if self.path_exists(path):
            context.log.warning(f"Removing existing ADLS2 key: {path}")  # noqa: G004
            self.unlink(path)

        file = self.file_system_client.create_file(path.as_posix())
        with self._acquire_lease(file) as lease:
            # Note: chunk_size can also be specified, see API for Azure SDK for Python, DataLakeFileClient:
            # https://learn.microsoft.com/en-us/python/api/azure-storage-file-datalake/azure.storage.filedatalake.datalakefileclient
            file.upload_data(
                obj,
                lease=lease,
                overwrite=True,
                connection_timeout=_CONNECTION_TIMEOUT,
            )

    def path_exists(self, path: UPath) -> bool:
        try:
            self.file_system_client.get_file_client(
                path.as_posix()
            ).get_file_properties()
        except ResourceNotFoundError:
            return False
        return True

    @contextmanager
    def _acquire_lease(self, client: Any, is_rm: bool = False) -> Iterator[str]:
        lease_client = self.lease_client_constructor(client=client)
        try:
            lease_client.acquire(lease_duration=self.lease_duration)
            yield lease_client.id
        finally:
            # cannot release a lease on a file that no longer exists, so need to check
            if not is_rm:
                lease_client.release()

    def handle_df(self, context: OutputContext, df: pd.DataFrame, path: UPath) -> None:
        self.dump_to_path(context, df.to_parquet(None), path)


class AzureMetadataIOManager(AzureMixin, MetadataIOManager):
    pass


class AzureGeoIOManager(AzureMixin, GeoIOManager):
    def handle_flatgeobuf(
        self, context: OutputContext, geo_df: gpd.GeoDataFrame, full_path: UPath
    ) -> None:
        tmp = tempfile.NamedTemporaryFile()
        fname = tmp.name + ".fgb"
        geo_df.to_file(fname, driver="FlatGeobuf")
        with Path(fname).open(mode="rb") as f:
            b: bytes = f.read()
            context.log.debug(ic(f"Size: {len(b) / (1_024 * 1_024):.3f}MB"))
            self.dump_to_path(context, b, full_path)

    def handle_geojsonseq(
        self, context: OutputContext, geo_df: gpd.GeoDataFrame, full_path: UPath
    ) -> None:
        tmp = tempfile.NamedTemporaryFile()
        fname = tmp.name + ".geojsonseq"
        geo_df.to_file(fname, driver="GeoJSONSeq")
        with Path(fname).open(mode="rb") as f:
            b: bytes = f.read()
            context.log.debug(ic(f"Size: {len(b) / (1_024 * 1_024):.3f}MB"))
            self.dump_to_path(context, b, full_path)

    def handle_names(
        self, context: OutputContext, names_df: pd.DataFrame, full_path: UPath
    ) -> None:
        self.dump_to_path(context, names_df.to_parquet(None), full_path)

    def handle_geo_metadata(
        self, context: OutputContext, geo_metadata_df: pd.DataFrame, full_path: UPath
    ) -> None:
        self.dump_to_path(context, geo_metadata_df.to_parquet(None), full_path)


class AzureMetricsIOManager(AzureMixin, MetricsIOManager):
    pass


class AzureMetricsPartitionedIOManager(AzureMixin, MetricsPartitionedIOManager):
    pass


class AzureMetricsMetadataIOManager(AzureMixin, MetricsMetdataIOManager):
    pass


class AzureGeneralIOManager(AzureMixin, IOManager):
    """This class is used only for an asset which tests the Azure functionality
    (see cloud_outputs/azure_test.py). It is not used for publishing any
    poppusher data."""

    extension: str

    def __init__(self, extension: str | None = None):
        super().__init__()
        if extension is not None and not extension.startswith("."):
            err_msg = f"Provided extension ('{extension}') does not begin with '.'"
            raise ValueError(err_msg)
        self.extension = "" if extension is None else extension

    def handle_output(self, context: OutputContext, obj: bytes) -> None:
        path = self.get_base_path() / ".".join(
            [*context.asset_key.path, self.extension]
        )
        self.dump_to_path(context, obj, path)

    def load_input(self, context: InputContext) -> Any:
        return super().load_input(context)



from __future__ import annotations

import base64
import datetime
import io
import subprocess
import xml.etree.ElementTree as ET
import zipfile
from io import BytesIO
from pathlib import Path
from tempfile import TemporaryDirectory

import fsspec
import geopandas as gpd
import matplotlib.pyplot as plt
import pandas as pd
import requests
from dagster import (
    ConfigurableResource,
    EnvVar,
    MetadataValue,
    get_dagster_logger,
)

DOWNLOAD_ROOT = Path(__file__).parent.absolute() / "data"
CACHE_ROOT = Path(__file__).parent.absolute() / "cache"


class SourceDataAssumptionsOutdated(ValueError):
    """
    Raised when a DAG detected a situation that implies there has been a change to
    the assumed properties of the upstream data.
    Typically this error implies that the DAG will need to be retested and updated.
    """


class StagingDirResource(ConfigurableResource):
    staging_dir: str | None


def markdown_from_plot() -> str:
    plt.tight_layout()
    # Convert the image to a saveable format
    buffer = BytesIO()
    plt.savefig(buffer, format="png")
    image_data = base64.b64encode(buffer.getvalue())

    # Convert the image to Markdown to preview it within Dagster
    return f"![img](data:image/png;base64,{image_data.decode()})"


def _last_update(file_path):
    """
    Returns the date and time of the last update to the file at `file_path`.
    """
    _path = Path(file_path)
    if not _path.exists():
        return None
    last_update = _path.stat().st_mtime
    return datetime.datetime.fromtimestamp(last_update)


def get_staging_dir(
    asset_context, staging_res: StagingDirResource
) -> StagingDirectory | TemporaryDirectory:
    """
    This creates a "staging directory" resource. This is a convenience feature for developers,
    allowing large downloads to be cached locally and not repeatedly downloaded during development.

    - The root of the staging directory is specified by the `staging_dir` resource.
    - If staging_dir is `None` then a [TemporaryDirectory](https://docs.python.org/3/library/tempfile.html#tempfile.TemporaryDirectory)
    will be returned.
    - If staging_dir is not specified, a ValueError is raised.

    The staging directory must be separate from, and must _not_ contained within, the Dagster-managed
    asset storage (as specified by DAGSTER_HOME). A ValueError will be raised if this is not the case.

    Internally, it is organised using asset-prefix / asset-name / [partition-name] to ensure that
    each asset has an (initially) empty directory with which to work.Typically the directory returned
    will have the path structure:

    ```
    staging_dir_root / asset_prefix / asset_name / [partition_name]
    ```

    If called from a multi-asset function, the directory will have the path structure:

    ```
    staging_dir_root / multi_asset_function_name
    ```

    There is no cache validation/invalidation logic (nor is any planned). The developer is responsible
    for clearing/deleting its contents when necessary.

    The "staging directory" should typically be used where TemporaryDirectory might be used.
    The "staging directory" is not intended for use in production deployments.
    """
    if staging_res.staging_dir is None:
        return TemporaryDirectory()

    return StagingDirectory(asset_context, staging_res.staging_dir)


class StagingDirectory:
    def __init__(self, asset_context, staging_root: str):
        self.staging_dir: Path = self._get_staging_dir(asset_context, staging_root)
        self._check_against_dagster_home()

    def _check_against_dagster_home(self):
        # Check that the staging directory is not the same as, or contained within DAGSTER_HOME
        dagster_home_str = EnvVar("DAGSTER_HOME").get_value(None)
        if dagster_home_str:
            dagster_home = Path(dagster_home_str)
            dagster_home = dagster_home.resolve()

            if dagster_home == self.staging_dir:
                err_msg = "DAGSTER_HOME and staging_dir are the same"
                raise ValueError(err_msg)

            if dagster_home in self.staging_dir.parents:
                err_msg = "DAGSTER_HOME is a parent of staging_dir"
                raise ValueError(err_msg)

    # TODO: should this be a class method or instance method?
    def _get_staging_dir(self, asset_context, staging_root: str) -> Path:
        root = Path(staging_root)
        root = root.resolve()

        if asset_context.op_config and "unit_test_key" in asset_context.op_config:
            # TODO: This is unsatisfactory, as it is necessary to have code that handle the specific
            # conditions of running in a test environment. The `context` object in a test
            # context is not a perfect replica of the `context` object. There are many
            # attributes that are not present in the test context.
            #
            # Any attempt to access the context properties, including checking for them
            # using `hasattr` throws an `dagster._core.errors.DagsterInvalidPropertyError`
            # error. This is not caught by either `DagsterError` or `DagsterUserCodeExecutionError`.
            get_dagster_logger().info(
                "This is unsatisfactory, as it is necessary to have code that handle the specific"
                " conditions of running in a test environment. The `context` object in a test"
                " context is not a perfect replica of the `context` object. There are many "
                " attributes that are not present in the test context. "
                " Any attempt to access the context properties, including checking for them"
                " using `hasattr` throws an `dagster._core.errors.DagsterInvalidPropertyError`"
                " error. This is not caught by either `DagsterError` or `DagsterUserCodeExecutionError`."
            )

            staging_dir = root / asset_context.op_config["unit_test_key"]
        elif len(asset_context.selected_asset_keys) == 1:
            asset_key = asset_context.asset_key_for_output()
            staging_dir = root.joinpath(*asset_key.path)
        else:
            # Assuming that this is being called from a function that returns
            # a multi_asset, hence there is no single asset key to use
            # `asset_context.op_handle.name` is the name of the multi_asset function
            multi_asset_func_name = asset_context.op_handle.name
            staging_dir = root.joinpath(multi_asset_func_name)

        # Add in the partition key if it exists
        if asset_context.has_partition_key:
            partition_key = asset_context.partition_key
            staging_dir = staging_dir / partition_key

        return staging_dir

    def __enter__(self) -> str:
        # Ensure staging_dir exists
        self.staging_dir.mkdir(parents=True, exist_ok=True)
        get_dagster_logger().info(f"Using staging directory: {self.staging_dir}")
        return str(self.staging_dir)

    def __exit__(self, exc_type, exc_value, traceback):
        get_dagster_logger().info(f"Exiting staging directory: {self.staging_dir}")


def extract_main_file_from_zip(temp_zip_file, temp_dir, expected_extension) -> str:
    """
    For cases where the downloaded file (a) is a zip file, (b) contains multiple files,
    and (c) we only want one of those files, (e.g the other files are metadata or documentation
    files), this function extracts the main file from a zip file.

    Parameters:
    -----------
    temp_file : str
        The path to the downloaded file.
    temp_dir : str
        The path to the directory where the file will be extracted.
    expected_extension : str
        The file extension of the file we want to extract. It is assumed that there will only be
        one file with this extension in the zip file.
    """
    source_archive_file_path = expected_extension
    # We have a zip file, so extract the file we want from that
    with zipfile.ZipFile(temp_zip_file, "r") as z:
        # if there is only one file in the zip, we can just extract it
        zip_contents = z.namelist()

        # If there is only one file in the zip, we can just extract it, ignoring the specified file path
        if len(zip_contents) == 1:
            source_archive_file_path = zip_contents[0]
        elif sum([f.endswith(expected_extension) for f in zip_contents]) == 1:
            # If there are multiple files in the zip, but only one has the correct file extension, we can just extract it, ignoring the specified file path
            for f in zip_contents:
                if f.endswith(expected_extension):
                    source_archive_file_path = f

        # Extract the file we want, assuming that we've found the identified the correct file
        if source_archive_file_path in zip_contents:
            z.extract(source_archive_file_path, path=temp_dir)
            return str(temp_dir / source_archive_file_path)

        # If we get here, we have not found the file we want
        err_msg = (
            f"Could not find {source_archive_file_path} in the zip file {temp_zip_file}"
        )
        raise ValueError(err_msg)


def download_from_wfs(wfs_url: str, output_file: str) -> None:
    """
    Downloads data from a WFS (`wfs_url`) and saves it to a file (`output_file`).

    The `ogr2ogr` command line tool is used to workaround the feature count limit that can be imposed by the server. (See https://gdal.org/drivers/vector/wfs.html#request-paging and https://gis.stackexchange.com/questions/422609/downloading-lots-of-data-from-wfs for details).

    Parameters
    ----------
    wfs_url : str
        The URL of the WFS.
    output_file : str
        The name of the output file.
    """
    template = r"""
        <OGRWFSDataSource>
            <URL>CHANGE_ME</URL>
            <PagingAllowed>ON</PagingAllowed>
            <PageSize>1000</PageSize>
        </OGRWFSDataSource>
    """

    # Writing OGR Virtual Format file
    root = ET.fromstring(
        template
    )  # This will return None is the XML template is invalid, or and Element if it is valid.
    root.find("URL").text = wfs_url  # type: ignore (In this case we can be confident that the template is valid)

    with Path(f"{output_file}.xml").open(mode="w") as f:
        f.write(ET.tostring(root).decode())

    # Running ogr2ogr
    # *Assuming* that ogr2ogr returns zero on success (have not confirmed this independently)
    subprocess.run(
        ["ogr2ogr", "-f", "GeoJSON", f"{output_file}.geojson", f"{output_file}.xml"],
        check=True,
    )

    # Done


def get_path_to_cache(
    url: str, cache_path: Path, mode: str = "b"
) -> fsspec.core.OpenFiles:
    """
    Returns the path(s) to the local cached files for a given URL.
    Downloads the file if it is not already cached.
    """
    cache_storage = str(CACHE_ROOT / cache_path)

    # If a cache path is provided, enforce the use of simplecache
    if cache_path and not url.startswith("simplecache::"):
        url = f"simplecache::{url}"

    if ("w" not in mode) and ("r" not in mode):
        mode = f"r{mode}"

    return fsspec.open(
        url,
        mode,
        simplecache={
            "cache_storage": cache_storage,
            "check_files": False,
        },
    )


def download_zipped_files(zipfile_url: str, output_dir: str) -> None:
    """
    Downloads a zipped file from a URL and extracts it to a folder.

    Raises a ValueError if the output folder is not empty.
    """
    output_dpath = Path(output_dir).absolute()
    output_dpath.mkdir(parents=True, exist_ok=True)

    if any(output_dpath.iterdir()):
        err_msg = f"Directory {output_dpath} is not empty. Skipping download."
        raise ValueError(err_msg)

    r = requests.get(zipfile_url)
    z = zipfile.ZipFile(io.BytesIO(r.content))
    z.extractall(output_dpath)


def add_metadata(
    context,
    df: pd.DataFrame | gpd.GeoDataFrame,
    title: str | list[str],
    output_name: str | None = None,
):
    context.add_output_metadata(
        metadata={
            "title": title,
            "num_records": len(df),
            "columns": MetadataValue.md(
                "\n".join([f"- '`{col}`'" for col in df.columns.to_list()])
            ),
            "preview": MetadataValue.md(df.head().to_markdown()),
        },
        output_name=output_name,
    )


if __name__ == "__main__":
    pass
    # This is for testing only
    # oa_wfs_url: str = "https://dservices1.arcgis.com/ESMARspQHYMw9BZ9/arcgis/services/Output_Areas_Dec_2021_Boundaries_Generalised_Clipped_EW_BGC_/WFSServer?service=wfs",
    # layer_name: str = "Output_Areas_Dec_2021_Boundaries_Generalised_Clipped_EW"

    # serviceItemId taken from:
    # https://services1.arcgis.com/ESMARspQHYMw9BZ9/arcgis/rest/services/Output_Areas_Dec_2021_Boundaries_Generalised_Clipped_EW_BGC_2022/FeatureServer/0?f=pjson
    # serviceItemId: str = "6c6743e1e4b444f6afcab9d9588f5d8f"

    # download_from_wfs(oa_wfs_url, layer_name)
    # download_from_arcgis_online(serviceItemId, "data/oa_from_agol2.geojson")



from __future__ import annotations

import pytest

# generate_pmtiles,
# from poppusher.cloud_outputs import (
#    cartography_in_cloud_formats,
#    country_outputs_sensor,
# )
# generate_pmtiles,
# TODO, Move this to a fixture to somewhere more universal
from .test_bel import demo_sectors  # noqa: F401

# Commented out test as part of #92 as functions no longer importable
# @pytest.mark.skip(
#     reason="This tests old code that is no longer imported since cloud outputs were reworked in #92"
# )
# def test_country_outputs_sensor():
#     """ ""
#     This test checks that the country_outputs_sensor function returns a RunRequest
#     with the correct AssetKey and PartitionKey.

#     Three upstream assets are mocked, which different qualities.
#     The context object is mocked and this also defines which assets are monitored. Hence the test does not attempt to
#     check that the sensor is correctly monitoring the required production assets.

#     Hence this test only checks that the correct config parameters are returned via the RunRequest object(s).
#     """

#     @asset
#     def irrelevant_asset():
#         return 1

#     @asset
#     def monitored_asset():
#         return 2

#     @asset(
#         partitions_def=StaticPartitionsDefinition(["A", "B", "C"]),
#     )
#     def partitioned_monitored_asset(context):
#         return ic(context.partition_key)

#     # instance = DagsterInstance.ephemeral()
#     with instance_for_test() as instance:
#         ctx = build_multi_asset_sensor_context(
#             monitored_assets=[
#                 AssetKey("monitored_asset"),
#                 AssetKey("partitioned_monitored_asset"),
#             ],
#             instance=instance,
#             definitions=defs,
#         )

#         # Nothing is materialized yet
#         result = list(country_outputs_sensor(ctx))
#         assert len(result) == 1
#         assert isinstance(result[0], SkipReason)

#         # Only the irrelevant asset is materialized
#         materialize_to_memory([irrelevant_asset], instance=instance)
#         result = list(country_outputs_sensor(ctx))
#         assert len(result) == 1
#         assert isinstance(result[0], SkipReason)

#         # Only the monitored asset is materialized
#         materialize_to_memory([monitored_asset], instance=instance)
#         result = list(country_outputs_sensor(ctx))
#         assert len(result) == 1
#         assert isinstance(result[0], RunRequest)

#         # Both non-partitioned assets are materialized
#         materialize_to_memory([monitored_asset, irrelevant_asset], instance=instance)
#         result = list(country_outputs_sensor(ctx))
#         assert len(result) == 1
#         assert isinstance(result[0], RunRequest)
#         assert result[0].partition_key == "monitored-asset"
#         assert (
#             result[0].run_config["ops"]["upstream_df"]["config"]["asset_to_load"]
#             == "monitored_asset"
#         )
#         assert (
#             result[0].run_config["ops"]["upstream_df"]["config"]["partition_to_load"]
#             is None
#         )

#         # All three assets are materialized
#         materialize_to_memory(
#             [monitored_asset, irrelevant_asset, partitioned_monitored_asset],
#             instance=instance,
#             partition_key="A",
#         )
#         result = list(country_outputs_sensor(ctx))
#         assert len(result) == 2

#         # The order of the results does not need to be guaranteed
#         # These two functions are used to check the results below
#         def assert_for_non_partitioned_assets(r):
#             assert isinstance(r, RunRequest)
#             assert r.partition_key == "monitored-asset"
#             assert (
#                 r.run_config["ops"]["upstream_df"]["config"]["asset_to_load"]
#                 == "monitored_asset"
#             )
#             assert (
#                 r.run_config["ops"]["upstream_df"]["config"]["partition_to_load"]
#                 is None
#             )

#         def assert_for_partitioned_assets(r):
#             assert r.partition_key == "partitioned-monitored-asset-a"
#             assert (
#                 r.run_config["ops"]["upstream_df"]["config"]["asset_to_load"]
#                 == "partitioned_monitored_asset"
#             )
#             assert (
#                 r.run_config["ops"]["upstream_df"]["config"]["partition_to_load"] == "A"
#             )

#         # Now check that the results include one partitioned and one non-partitioned asset
#         try:
#             assert_for_non_partitioned_assets(result[0])
#             assert_for_partitioned_assets(result[1])
#         except AssertionError:
#             assert_for_non_partitioned_assets(result[1])
#             assert_for_partitioned_assets(result[0])

# Commented out test as part of #92 as functions no longer importable
# @pytest.mark.skip(
#     reason="This tests old code that is no longer imported since cloud outputs were reworked in #92"
# )
# # TODO: The no QA comment below is pending moving the fixture to a more
# # universal location
# def test_cartography_in_cloud_formats(tmp_path, demo_sectors):
#     # This test is outputs each of the cartography outputs to individual files
#     # in the staging directory. The test then checks that the files exist and
#     # that they were created after the test started.
#     time_at_start = datetime.now()

#     staging_res = StagingDirResource(staging_dir=str(tmp_path))
#     resources_for_test = {
#         "staging_res": staging_res,
#         "unit_test_key": "test_cartography_in_cloud_formats",
#     }

#     with (
#         instance_for_test() as instance,
#         build_asset_context(
#             resources=resources_for_test,
#             instance=instance,
#             partition_key="historic-european-region",
#         ) as context,
#     ):
#         # Collect the results
#         # Results should be a generator which produces three Output objects
#         results = cartography_in_cloud_formats(context, demo_sectors)

#         output_paths = [r.value for r in list(results)]
#         # There should be 3 outputs
#         assert len(output_paths) == 3

#         # Check that the output paths exist and were created after the test started
#         for output_path in output_paths:
#             assert output_path.exists()
#             assert output_path.stat().st_mtime > time_at_start.timestamp()


@pytest.mark.skip(reason="Test not implemented yet")
def test_generate_pmtiles():
    # input_geojson_path = str(
    #     Path(__file__).parent / "demo_data" / "be_demo_sector.geojson"
    # )

    # with build_asset_context(
    #     partition_key="historic-european-region") as context:
    #     generate_pmtiles(context, input_geojson_path)

    # Assert something about the logs exists
    # Assert some output files exist

    pytest.fail("Test not implemented")



from __future__ import annotations

from pathlib import Path

import requests

HEADERS = {
    "User-Agent": "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:92.0) Gecko/20100101 Firefox/92.0"
}


def download_file(
    cache_dir: str,
    url: str,
    file_name: Path | None = None,
    headers: dict[str, str] = HEADERS,
) -> Path:
    """Downloads file checking first if exists in cache, returning file name."""
    file_name = Path(cache_dir) / url.split("/")[-1] if file_name is None else file_name
    if not Path(file_name).exists():
        r = requests.get(url, allow_redirects=True, headers=headers)
        with Path(file_name).open("wb") as fp:
            fp.write(r.content)
    return file_name



from __future__ import annotations

import pandas as pd
from dagster import asset


@asset(io_manager_key="azure_general_io_manager")
def test_azure():
    return pd.DataFrame({"col1": [1, 2], "col2": [3, 4]}).to_parquet(None)


@asset(io_manager_key="azure_general_io_manager")
def test_azure_large():
    return b"0" * (450 * 1024 * 1024 + 100)



from __future__ import annotations

from collections.abc import Sequence
from datetime import date
from enum import Enum
from hashlib import sha256
from typing import Self

import jcs
import pandas as pd
from pydantic import BaseModel, Field, computed_field, model_validator


class COL(Enum):
    """
    This class stores the eventual column names of all the metadata classes,
    which are used when serialising the metadata to a dataframe.
    """

    GEO_ID = "GEO_ID"

    COUNTRY_ID = "country_id"
    COUNTRY_NAME_SHORT_EN = "country_name_short_en"
    COUNTRY_NAME_OFFICIAL = "country_name_official"
    COUNTRY_ISO3 = "country_iso3"
    COUNTRY_ISO2 = "country_iso2"
    COUNTRY_ISO3166_2 = "country_iso3166_2"

    DATA_PUBLISHER_ID = "data_publisher_id"
    DATA_PUBLISHER_NAME = "data_publisher_name"
    DATA_PUBLISHER_URL = "data_publisher_url"
    DATA_PUBLISHER_DESCRIPTION = "data_publisher_description"
    DATA_PUBLISHER_COUNTRIES_OF_INTEREST = "data_publisher_countries_of_interest"

    GEOMETRY_ID = "geometry_id"
    GEOMETRY_FILEPATH_STEM = "geometry_filepath_stem"
    GEOMETRY_VALIDITY_PERIOD_START = "geometry_validity_period_start"
    GEOMETRY_VALIDITY_PERIOD_END = "geometry_validity_period_end"
    GEOMETRY_LEVEL = "geometry_level"
    GEOMETRY_HXL_TAG = "geometry_hxl_tag"

    SOURCE_DATA_RELEASE_ID = "source_data_release_id"
    SOURCE_DATA_RELEASE_NAME = "source_data_release_name"
    SOURCE_DATA_RELEASE_DATE_PUBLISHED = "source_data_release_date_published"
    SOURCE_DATA_RELEASE_REFERENCE_PERIOD_START = (
        "source_data_release_reference_period_start"
    )
    SOURCE_DATA_RELEASE_REFERENCE_PERIOD_END = (
        "source_data_release_reference_period_end"
    )
    SOURCE_DATA_RELEASE_COLLECTION_PERIOD_START = (
        "source_data_release_collection_period_start"
    )
    SOURCE_DATA_RELEASE_COLLECTION_PERIOD_END = (
        "source_data_release_collection_period_end"
    )
    SOURCE_DATA_RELEASE_EXPECT_NEXT_UPDATE = "source_data_release_expect_next_update"
    SOURCE_DATA_RELEASE_URL = "source_data_release_url"
    SOURCE_DATA_RELEASE_DATA_PUBLISHER_ID = "source_data_release_data_publisher_id"
    SOURCE_DATA_RELEASE_DESCRIPTION = "source_data_release_description"
    SOURCE_DATA_RELEASE_GEOMETRY_METADATA_ID = (
        "source_data_release_geometry_metadata_id"
    )

    METRIC_ID = "metric_id"
    METRIC_HUMAN_READABLE_NAME = "metric_human_readable_name"
    METRIC_SOURCE_METRIC_ID = "metric_source_id"
    METRIC_DESCRIPTION = "metric_description"
    METRIC_HXL_TAG = "metric_hxl_tag"
    METRIC_PARQUET_PATH = "metric_parquet_path"
    METRIC_PARQUET_COLUMN_NAME = "metric_parquet_column_name"
    METRIC_PARQUET_MARGIN_OF_ERROR_COLUMN = "metric_parquet_margin_of_error_column"
    METRIC_PARQUET_MARGIN_OF_ERROR_FILE = "metric_parquet_margin_of_error_file"
    METRIC_POTENTIAL_DENOMINATOR_IDS = "metric_potential_denominator_ids"
    METRIC_PARENT_METRIC_ID = "metric_parent_metric_id"
    METRIC_SOURCE_DATA_RELEASE_ID = "metric_source_data_release_id"
    METRIC_SOURCE_DOWNLOAD_URL = "metric_source_download_url"
    METRIC_SOURCE_ARCHIVE_FILE_PATH = "metric_source_archive_file_path"
    METRIC_SOURCE_DOCUMENTATION_URL = "metric_source_documentation_url"


class MetadataBaseModel(BaseModel):
    def hash_class_vars(self):
        """
        Calculate a SHA256 hash from a class instance's variables. Used for
        generating unique and verifiable IDs for metadata classes.

        Note that `vars()` does not include properties, so the IDs themselves are
        not part of the hash, which avoids self-reference issues.

        Furthermore, note that serialisation aliases are not used: the hash is
        calculated from the original field names.
        """

        # Must copy the dict to avoid overriding the actual instance attributes!
        # Because we're only modifying dates -> strings, we don't need to perform a
        # deepcopy but all variables must be serializable
        def serializable_vars(obj: object) -> dict:
            variables = {}
            # Check if variables are serializable
            for key, val in vars(obj).items():
                try:
                    # Python doesn't serialise dates to JSON, have to convert to ISO 8601 first
                    new_val = val.isoformat() if isinstance(val, date) else val
                    # Try to serialise
                    jcs.canonicalize(new_val)
                    # Store in dict if serialisable
                    variables[key] = new_val
                except Exception:
                    # If cannot serialise, continue
                    continue

            return variables

        return sha256(jcs.canonicalize(serializable_vars(self))).hexdigest()

    @classmethod
    def fix_types(cls, df: pd.DataFrame) -> pd.DataFrame:
        """
        When a list of MetadataBaseModel classes is converted to a dataframe,
        the types of the fields can be lost (notably, if all the values in a
        column are None). This method is implemented on each class in order to
        coerce the output dataframe columns to the correct types, so that the
        subsequent serialisation to parquet will retain this type information.

        In general, this function only needs to coerce columns which can be x
        or None to the type x itself --- for example, `iso3166_2` on the
        `CountryMetadata` class. For a column which is already a string (e.g.
        `iso3`), it should not be possible to instantiate the class with a None
        value, as Pydantic will raise an error.

        The default implementation of this method is to do nothing and just
        return the input dataframe. If a class needs to coerce types, it should
        override this method and return the modified dataframe.
        """
        return df


def metadata_to_dataframe(
    metadata_instances: Sequence[MetadataBaseModel],
):
    """
    Convert a list of metadata instances to a pandas DataFrame. Any of the five
    metadata classes defined in this module can be used here.
    """
    cls = metadata_instances[0].__class__
    return cls.fix_types(
        pd.DataFrame([md.model_dump(by_alias=True) for md in metadata_instances])
    )


class CountryMetadata(MetadataBaseModel):
    @computed_field(alias=COL.COUNTRY_ID.value)
    @property
    def id(self) -> str:
        if self.iso3166_2 is not None:
            return self.iso3166_2.lower().replace("-", "_")
        return self.iso3.lower().replace("-", "_")

    name_short_en: str = Field(
        description="The short name of the country in English (for example 'Belgium').",
        serialization_alias=COL.COUNTRY_NAME_SHORT_EN.value,
    )
    name_official: str = Field(
        description="The official name of the country (for example 'Kingdom of Belgium'). In English if available.",
        serialization_alias=COL.COUNTRY_NAME_OFFICIAL.value,
    )
    iso3: str = Field(
        description="The ISO 3166-1 alpha-3 code of the country (for example 'BEL').",
        serialization_alias=COL.COUNTRY_ISO3.value,
    )
    iso2: str = Field(
        description="The ISO 3166-1 alpha-2 code of the country (for example 'BE').",
        serialization_alias=COL.COUNTRY_ISO2.value,
    )
    iso3166_2: str | None = Field(
        description="If the territory is a 'principal subdivision', its ISO 3166-2 code (for example 'BE-VLG').",
        serialization_alias=COL.COUNTRY_ISO3166_2.value,
    )

    @classmethod
    def fix_types(cls, df: pd.DataFrame) -> pd.DataFrame:
        return df.astype(
            {
                COL.COUNTRY_ISO3166_2.value: "string",
            }
        )


class DataPublisher(MetadataBaseModel):
    @computed_field(alias=COL.DATA_PUBLISHER_ID.value)
    @property
    def id(self) -> str:
        return self.hash_class_vars()

    name: str = Field(
        description="The name of the organisation publishing the data",
        serialization_alias=COL.DATA_PUBLISHER_NAME.value,
    )
    url: str = Field(
        description="The URL of the publisher's homepage.",
        serialization_alias=COL.DATA_PUBLISHER_URL.value,
    )
    description: str = Field(
        description="A brief description of the organisation publishing the data, including its mandate.",
        serialization_alias=COL.DATA_PUBLISHER_DESCRIPTION.value,
    )
    countries_of_interest: list[str] = Field(
        description="A list of country IDs for which the publisher has data available.",
        serialization_alias=COL.DATA_PUBLISHER_COUNTRIES_OF_INTEREST.value,
    )


class GeometryMetadata(MetadataBaseModel):
    @computed_field(alias=COL.GEOMETRY_ID.value)
    @property
    def id(self) -> str:
        return self.hash_class_vars()

    @computed_field(alias=COL.GEOMETRY_FILEPATH_STEM.value)
    @property
    def filepath_stem(self) -> str:
        level = "_".join(self.level.lower().split())
        year = self.validity_period_start.year
        return f"{self.country_metadata.id}/geometries/{level}_{year}"

    country_metadata: CountryMetadata = Field(
        "The `CountryMetadata` associated with the geometry.", exclude=True
    )

    validity_period_start: date = Field(
        description="The start of the range of time for which the regions are valid (inclusive)",
        serialization_alias=COL.GEOMETRY_VALIDITY_PERIOD_START.value,
    )
    validity_period_end: date = Field(
        description="The end of the range of time for which the regions are valid (inclusive). If the data is a single-day snapshot, this should be the same as `validity_period_start`.",
        serialization_alias=COL.GEOMETRY_VALIDITY_PERIOD_END.value,
    )
    level: str = Field(
        description="The geography level contained in the file (e.g. output area, LSOA, MSOA, etc)",
        serialization_alias=COL.GEOMETRY_LEVEL.value,
    )
    hxl_tag: str = Field(
        description="Humanitarian eXchange Language (HXL) description for the geography level",
        serialization_alias=COL.GEOMETRY_HXL_TAG.value,
    )


class SourceDataRelease(MetadataBaseModel):
    @computed_field(alias=COL.SOURCE_DATA_RELEASE_ID.value)
    @property
    def id(self) -> str:
        return self.hash_class_vars()

    name: str = Field(
        description="The name of the data release, as given by the publisher",
        serialization_alias=COL.SOURCE_DATA_RELEASE_NAME.value,
    )
    date_published: date = Field(
        description="The date on which the data was published",
        serialization_alias=COL.SOURCE_DATA_RELEASE_DATE_PUBLISHED.value,
    )
    reference_period_start: date = Field(
        description="The start of the range of time for which the data can be assumed to be valid (inclusive)",
        serialization_alias=COL.SOURCE_DATA_RELEASE_REFERENCE_PERIOD_START.value,
    )
    reference_period_end: date = Field(
        description="The end of the range of time for which the data can be assumed to be valid (inclusive). If the data is a single-day snapshot, this should be the same as `reference_period_start`.",
        serialization_alias=COL.SOURCE_DATA_RELEASE_REFERENCE_PERIOD_END.value,
    )
    collection_period_start: date = Field(
        description="The start of the range of time during which the data was collected (inclusive)",
        serialization_alias=COL.SOURCE_DATA_RELEASE_COLLECTION_PERIOD_START.value,
    )
    collection_period_end: date = Field(
        description="The end of the range of time during which the data was collected (inclusive). If the data were collected in a single day, this should be the same as `collection_period_start`.",
        serialization_alias=COL.SOURCE_DATA_RELEASE_COLLECTION_PERIOD_END.value,
    )
    expect_next_update: date = Field(
        description="The date on which is it expected that an updated edition of the data will be published. In some cases this will be the same as `reference_period_end`",
        serialization_alias=COL.SOURCE_DATA_RELEASE_EXPECT_NEXT_UPDATE.value,
    )
    url: str = Field(
        description="The url of the data release.",
        serialization_alias=COL.SOURCE_DATA_RELEASE_URL.value,
    )
    data_publisher_id: str = Field(
        description="The ID of the publisher of the data release",
        serialization_alias=COL.SOURCE_DATA_RELEASE_DATA_PUBLISHER_ID.value,
    )
    description: str = Field(
        description="A description of the data release",
        serialization_alias=COL.SOURCE_DATA_RELEASE_DESCRIPTION.value,
    )
    geometry_metadata_id: str = Field(
        description="The ID of the geometry metadata associated with this data release",
        serialization_alias=COL.SOURCE_DATA_RELEASE_GEOMETRY_METADATA_ID.value,
    )

    @model_validator(mode="after")
    def check_dates(self) -> Self:
        msg_template = "{s}_period_start must be before or equal to {s}_period_end"
        if self.reference_period_start > self.reference_period_end:
            error_msg = msg_template.format(s="reference")
            raise ValueError(error_msg)
        if self.collection_period_start > self.collection_period_end:
            error_msg = msg_template.format(s="collection")
            raise ValueError(error_msg)
        return self


class MetricMetadata(MetadataBaseModel):
    @computed_field(alias=COL.METRIC_ID.value)
    @property
    def id(self) -> str:
        return self.hash_class_vars()

    human_readable_name: str = Field(
        description='A human readable name for the metric, something like "Total Population under 12 years old"',
        serialization_alias=COL.METRIC_HUMAN_READABLE_NAME.value,
    )
    source_metric_id: str = Field(
        description='The name of the metric that comes from the source dataset (for example in the ACS this might be "B001_E001" or something similar)',
        serialization_alias=COL.METRIC_SOURCE_METRIC_ID.value,
    )
    description: str = Field(
        description="A longer description of the metric which might include info on the caveats for the metric",
        serialization_alias=COL.METRIC_DESCRIPTION.value,
    )
    hxl_tag: str = Field(
        description="Field description using the Humanitarian eXchange Language (HXL) standard",
        serialization_alias=COL.METRIC_HXL_TAG.value,
    )
    metric_parquet_path: str = Field(
        description="The path to the parquet file that contains the metric",
        serialization_alias=COL.METRIC_PARQUET_PATH.value,
    )
    parquet_column_name: str = Field(
        description="Name of column in the outputted parquet file which contains the metric",
        serialization_alias=COL.METRIC_PARQUET_COLUMN_NAME.value,
    )
    parquet_margin_of_error_column: str | None = Field(
        description="Name of the column if any that contains the margin of error for the metric",
        serialization_alias=COL.METRIC_PARQUET_MARGIN_OF_ERROR_COLUMN.value,
    )
    parquet_margin_of_error_file: str | None = Field(
        description="Location (url) of the parquet file that contains the margin of error for the metric",
        serialization_alias=COL.METRIC_PARQUET_MARGIN_OF_ERROR_FILE.value,
    )
    potential_denominator_ids: list[str] | None = Field(
        description="A list of metrics which are suitable denominators for this metric.",
        serialization_alias=COL.METRIC_POTENTIAL_DENOMINATOR_IDS.value,
    )
    parent_metric_id: str | None = Field(
        description="Metric if any which is the parent to this one ( some census data like the ACS is organised hierarchically, this can be useful for making the metadata more searchable)",
        serialization_alias=COL.METRIC_PARENT_METRIC_ID.value,
    )
    source_data_release_id: str = Field(
        description="The id of the data release from which this metric comes",
        serialization_alias=COL.METRIC_SOURCE_DATA_RELEASE_ID.value,
    )
    source_download_url: str = Field(
        description="The url used to download the data from source.",
        serialization_alias=COL.METRIC_SOURCE_DOWNLOAD_URL.value,
    )
    source_archive_file_path: str | None = Field(
        description="(Optional), If the downloaded data is in an archive file (eg zip, tar, etc), this field is the path with the archive to locate the data file.",
        serialization_alias=COL.METRIC_SOURCE_ARCHIVE_FILE_PATH.value,
    )
    source_documentation_url: str = Field(
        description="The documentation of the data release in human readable form.",
        serialization_alias=COL.METRIC_SOURCE_DOCUMENTATION_URL.value,
    )

    @classmethod
    def fix_types(cls, df: pd.DataFrame) -> pd.DataFrame:
        return df.astype(
            {
                COL.METRIC_PARQUET_MARGIN_OF_ERROR_COLUMN.value: "string",
                COL.METRIC_PARQUET_MARGIN_OF_ERROR_FILE.value: "string",
                COL.METRIC_POTENTIAL_DENOMINATOR_IDS.value: "object",
                COL.METRIC_PARENT_METRIC_ID.value: "string",
                COL.METRIC_SOURCE_ARCHIVE_FILE_PATH.value: "string",
            }
        )


EXPORTED_MODELS = [
    CountryMetadata,
    DataPublisher,
    SourceDataRelease,
    MetricMetadata,
    GeometryMetadata,
]


def export_schema():
    """
    Generates a JSON schema for all the models in this script and outputs it to
    the specified directory, with the filename `poppusher_{VERSION}.json`.
    """
    import argparse
    import json
    from pathlib import Path

    from pydantic.json_schema import models_json_schema

    from poppusher import __version__

    parser = argparse.ArgumentParser(description=export_schema.__doc__)
    parser.add_argument(
        "out_dir", help="The directory to output the schema to. Must exist."
    )
    args = parser.parse_args()
    out_dir = Path(args.out_dir)

    _, top_level_schema = models_json_schema(
        [(model, "serialization") for model in EXPORTED_MODELS],
        title="poppusher_schema",
        description=f"Version {__version__}",
    )
    if not out_dir.exists():
        error_msg = f"Directory {out_dir} does not exist."
        raise FileNotFoundError(error_msg)
    with (out_dir / f"poppusher_{__version__}.json").open("w") as f:
        json.dump(top_level_schema, f, indent=2)
